from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.losses import SigmoidCrossEntropy
import torch
import torch.nn as nn
from typing import Callable, Optional, List
from torch import Tensor


class IRVLayer(nn.Module):
    """
    Implements Influence Relevance Voter (IRV), a novel machine learning model designed for virtual
    high-throughput screening (vHTS).vHTS predicts the biological activity of chemical compounds
    sing computational methods, reducing the need for expensive experimental screening.

    The IRV model extends the k-Nearest Neighbors (kNN) algorithm by improving how neighbors
    influence predictions. Instead of treating all neighbors equally, IRV assigns each neighbor
    a relevance score based on its similarity to the query compound.This similarity is calculated
    using molecular fingerprint comparisons between the query compound and its neighbors,allowing
    more relevant neighbors to have a greater impact on the prediction.

    This model has been benchmarked on HIV dataset from IJCNN-07 Competition organised in 2007
    and DHFR dataset from McMaster University Data-Mining and Docking Competition organised in 2005 in [1].

    Example
    -------

    >>> import deepchem as dc
    >>> import numpy as np
    >>> n_tasks = 5
    >>> n_samples = 10
    >>> n_features = 128
    >>> K=5

    >>> # Generate dummy dataset.
    >>> ids = np.arange(n_samples)

    >>> # Features in ECFP Fingerprints representation
    >>> X = np.random.randint(2, size=(n_samples, n_features))

    >>> # Labels for tasks.
    >>> # Either 1 or 0 depending on whether the samples are active or inactive in that application(task)
    >>> y = np.ones((n_samples, n_tasks))

    >>> # Weights for each task in each column.
    >>> w = np.ones((n_samples, n_tasks))

    >>> dataset = dc.data.NumpyDataset(X, y, w, ids)

    >>> # Transforms ECFP Fingerprints to IRV features(Similarity values of top K nearest neighbors).

    >>> # Initialize the IRVTransformer with the reference dataset.
    >>> IRV_transformer = dc.trans.IRVTransformer(K, n_tasks, dataset)

    >>> # Apply the IRVTransformer.transform() to the target dataset for which the prediction is needed.
    >>> # Calculates the similrity values of the samples in target dataset with the reference dataset
    >>> # and returns the values of top K similar samples in reference dataset for each sample in target dataset.
    >>> dataset_trans = IRV_transformer.transform(dataset)

    >>> # Instantiate the model
    >>> model = dc.models.torch_models.MultitaskIRVClassifier(n_tasks, K = 5, learning_rate = 0.01, batch_size = n_samples)
    >>> # Train the model
    >>> loss = model.fit(dataset_trans)

    >>> # Prediction
    >>> output = model.predict(dataset_trans)

    >>> # Evaluation
    >>> classification_metric = dc.metrics.Metric(dc.metrics.accuracy_score,task_averager=np.mean)
    >>> score = model.evaluate(dataset_trans, [classification_metric])

    References:
    -----------
    [1] .. S.J.Swamidass et al,  "The Influence Relevance Voter: An Accurate and Interpretable Virtual High Throughput Screening Method.
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2750043/
    """

    def __init__(self, n_tasks: int, K, penalty: int):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks
        K: int
            Number of nearest neighbours used in classification
        penalty: float
            Amount of penalty (l2 or l1 applied)
        """
        super(IRVLayer, self).__init__()
        self.n_tasks = n_tasks
        self.K = K
        self.penalty = penalty

        # Initialize weights and biases
        self.V = nn.Parameter(torch.tensor([0.01, 1.], dtype=torch.float32))
        self.W = nn.Parameter(torch.tensor([1., 1.], dtype=torch.float32))
        self.b = nn.Parameter(torch.tensor([0.01], dtype=torch.float32))
        self.b2 = nn.Parameter(torch.tensor([0.01], dtype=torch.float32))

    def forward(self, inputs: Tensor) -> List[Tensor]:
        """Build the model."""

        K = self.K
        predictions = []
        for count in range(self.n_tasks):

            # Similarity values
            similarity = inputs[:, 2 * K * count:(2 * K * count + K)]

            # Labels for all top K similar samples
            ys = inputs[:, (2 * K * count + K):2 * K * (count + 1)].int()

            # Relevance
            R = self.b + self.W[0] * similarity + self.W[1] * torch.arange(
                1, K + 1, dtype=torch.float32, device=similarity.device)
            R = torch.sigmoid(R)

            # Influence = Relevance * Vote
            z = torch.sum(R * self.V[ys], dim=1) + self.b2
            predictions.append(z.view(-1, 1))

        logits = []
        outputs = []
        for task in range(self.n_tasks):
            task_output = Slice(task, 1)(torch.cat(predictions, dim=1))
            sigmoid = torch.sigmoid(task_output)
            logits.append(task_output)
            outputs.append(sigmoid)
        outputs_stacked = torch.stack(outputs, dim=1)
        outputs2 = 1 - outputs_stacked
        outputs = [
            torch.cat([outputs2, outputs_stacked], dim=2),
            logits[0] if len(logits) == 1 else torch.cat(logits, dim=1)
        ]
        return outputs


class Slice(nn.Module):
    """ Choose a slice of input on the last axis given order,
    Suppose input x has two dimensions,
    output f(x) = x[:, slice_num:slice_num+1]

    Extracts a specific slice (or "column/row/channel".)
    from a given input tensor along a specified dimension. Column is extracted
    from a 2D tensor here as the default value of axis is 1.
    -------

    >>> import deepchem as dc
    >>> import numpy as np
    >>> n_tasks = 5
    >>> n_samples = 10

    >>> # Generate dummy dataset.
    >>> predictions = torch.rand(n_samples, n_tasks)
    >>> # Extract slice
    >>> for task in range(n_tasks):
    ...     task_output = Slice(slice_num=task, axis=1)(predictions)

    """

    def __init__(self, slice_num: int, axis=1):
        """
        Parameters
        ----------
        slice_num: int
            index of slice number
        axis: int
            axis id
        """
        super(Slice, self).__init__()
        self.slice_num = slice_num
        self.axis = axis

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Returns a specific slice from the input tensor along the specified dimension."""
        slice_num = self.slice_num
        axis = self.axis
        device = inputs.device  # Get the device of the input tensor
        return inputs.index_select(axis, torch.tensor([slice_num],
                                                      device=device))


class MultitaskIRVClassifier(TorchModel):
    """
    Implements Influence Relevance Voter (IRV), a novel machine learning model designed for virtual
    high-throughput screening (vHTS).vHTS predicts the biological activity of chemical compounds
    sing computational methods, reducing the need for expensive experimental screening.

    The IRV model extends the k-Nearest Neighbors (kNN) algorithm by improving how neighbors
    influence predictions. Instead of treating all neighbors equally, IRV assigns each neighbor
    a relevance score based on its similarity to the query compound.This similarity is calculated
    using molecular fingerprint comparisons between the query compound and its neighbors,allowing
    more relevant neighbors to have a greater impact on the prediction.

    This model has been benchmarked on HIV dataset from IJCNN-07 Competition organised in 2007
    and DHFR dataset from McMaster University Data-Mining and Docking Competition organised in 2005 in [1].

    Example
    -------

    >>> import deepchem as dc
    >>> import numpy as np
    >>> n_tasks = 5
    >>> n_samples = 10
    >>> n_features = 128
    >>> K=5

    >>> # Generate dummy dataset.
    >>> ids = np.arange(n_samples)

    >>> # Features in ECFP Fingerprints representation
    >>> X = np.random.randint(2, size=(n_samples, n_features))

    >>> # Labels for tasks.
    >>> # Either 1 or 0 depending on whether the samples are active or inactive in that application(task)
    >>> y = np.ones((n_samples, n_tasks))

    >>> # Weights for each task in each column.
    >>> w = np.ones((n_samples, n_tasks))

    >>> dataset = dc.data.NumpyDataset(X, y, w, ids)

    >>> # Transforms ECFP Fingerprints to IRV features(Similarity values of top K nearest neighbors).

    >>> # Initialize the IRVTransformer with the reference dataset.
    >>> IRV_transformer = dc.trans.IRVTransformer(K, n_tasks, dataset)

    >>> # Apply the IRVTransformer.transform() to the target dataset for which the prediction is needed.
    >>> # Calculates the similrity values of the samples in target dataset with the reference dataset
    >>> # and returns the values of top K similar samples in reference dataset for each sample in target dataset.
    >>> dataset_trans = IRV_transformer.transform(dataset)

    >>> # Instantiate the model
    >>> model = dc.models.torch_models.MultitaskIRVClassifier(n_tasks, K = 5, learning_rate = 0.01, batch_size = n_samples)
    >>> # Train the model
    >>> loss = model.fit(dataset_trans)

    >>> # Prediction
    >>> output = model.predict(dataset_trans)

    >>> # Evaluation
    >>> classification_metric = dc.metrics.Metric(dc.metrics.accuracy_score,task_averager=np.mean)
    >>> score = model.evaluate(dataset_trans, [classification_metric])

    References:
    -----------
    [1] .. S.J.Swamidass et al,  "The Influence Relevance Voter: An Accurate and Interpretable Virtual High Throughput Screening Method.
    https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2750043/
    """

    def __init__(self,
                 n_tasks: int,
                 K=10,
                 penalty=0.0,
                 mode="classification",
                 device: Optional[torch.device] = None,
                 **kwargs):
        """Initialize MultitaskIRVClassifier

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        K: int
            Number of nearest neighbours used in classification
        penalty: float
            Amount of penalty (l2 or l1 applied)
        """

        self.n_tasks = n_tasks
        self.K = K
        self.n_features = 2 * self.K * self.n_tasks
        self.penalty = penalty

        # Define the IRVLayer
        self.model = IRVLayer(self.n_tasks, self.K, self.penalty)

        regularization_loss: Optional[Callable]
        if self.penalty != 0.0:
            weights = list(self.model.parameters())
            regularization_loss = lambda: self.penalty * torch.sum(  # noqa: E731
                torch.stack([torch.square(w).sum() for w in weights]))
        else:
            regularization_loss = None

        super(MultitaskIRVClassifier,
              self).__init__(self.model,
                             loss=SigmoidCrossEntropy(),
                             output_types=['prediction', 'loss'],
                             regularization_loss=regularization_loss,
                             **kwargs)



from deepchem.feat import Featurizer
from typing import List
try:
    from transformers import BertTokenizerFast
except ModuleNotFoundError:
    raise ImportError(
        'Transformers must be installed for BertFeaturizer to be used!')
    pass


class BertFeaturizer(Featurizer):
    """Bert Featurizer.

    Bert Featurizer.
    The Bert Featurizer is a wrapper class for HuggingFace's BertTokenizerFast.
    This class intends to allow users to use the BertTokenizer API while
    remaining inside the DeepChem ecosystem.

    Examples
    --------
    >>> from deepchem.feat import BertFeaturizer
    >>> from transformers import BertTokenizerFast
    >>> tokenizer = BertTokenizerFast.from_pretrained("Rostlab/prot_bert", do_lower_case=False)
    >>> featurizer = BertFeaturizer(tokenizer)
    >>> feats = featurizer.featurize(['D L I P [MASK] L V T'])

    Notes
    -----
    Examples are based on RostLab's ProtBert documentation.
    """

    def __init__(self, tokenizer: BertTokenizerFast):
        if not isinstance(tokenizer, BertTokenizerFast):
            raise TypeError(
                f"""`tokenizer` must be a constructed `BertTokenizerFast`
                object, not {type(tokenizer)}""")
        else:
            self.tokenizer = tokenizer

    def _featurize(self, datapoint: str, **kwargs) -> List[List[int]]:
        """
        Calculate encoding using HuggingFace's RobertaTokenizerFast

        Parameters
        ----------
        datapoint: str
            Arbitrary string sequence to be tokenized.

        Returns
        -------
        encoding: List
            List containing three lists: the `input_ids`, 'token_type_ids', and `attention_mask`.
        """

        # the encoding is natively a dictionary with keys 'input_ids', 'token_type_ids', and 'attention_mask'
        encoding = list(self.tokenizer(datapoint,
                                       **kwargs).values())  # type: ignore
        return encoding



# The number of elements to print for dataset ids/tasks
_print_threshold = 10


def get_print_threshold() -> int:
    """Return the printing threshold for datasets.

    The print threshold is the number of elements from ids/tasks to
    print when printing representations of `Dataset` objects.

    Returns
    ----------
    threshold: int
        Number of elements that will be printed
    """
    return _print_threshold


def set_print_threshold(threshold: int):
    """Set print threshold

    The print threshold is the number of elements from ids/tasks to
    print when printing representations of `Dataset` objects.

    Parameters
    ----------
    threshold: int
        Number of elements to print.
    """
    global _print_threshold
    _print_threshold = threshold


# If a dataset contains more than this number of elements, it won't
# print any dataset ids
_max_print_size = 1000


def get_max_print_size() -> int:
    """Return the max print size for a dataset.

    If a dataset is large, printing `self.ids` as part of a string
    representation can be very slow. This field controls the maximum
    size for a dataset before ids are no longer printed.

    Returns
    -------
    max_print_size: int
        Maximum length of a dataset for ids to be printed in string
        representation.
    """
    return _max_print_size


def set_max_print_size(max_print_size: int):
    """Set max_print_size

    If a dataset is large, printing `self.ids` as part of a string
    representation can be very slow. This field controls the maximum
    size for a dataset before ids are no longer printed.

    Parameters
    ----------
    max_print_size: int
        Maximum length of a dataset for ids to be printed in string
        representation.
    """
    global _max_print_size
    _max_print_size = max_print_size



from PIL import Image
import numpy as np

import time
import logging
from collections.abc import Sequence as SequenceCollection
from typing import Union, Dict, Tuple, Iterable, List, Optional, Callable, Any

from deepchem.models.optimizers import LearningRateSchedule
from deepchem.trans import Transformer, undo_transforms
from deepchem.utils.typing import LossFn, OneOrMany

import torch
from deepchem.models.torch_models.hf_models import HuggingFaceModel
from transformers import AutoProcessor, AutoModelForUniversalSegmentation, OneFormerConfig

logger = logging.getLogger(__name__)


class OneFormer(HuggingFaceModel):
    """
    Wrapper class that wraps the OneFormer model as a DeepChem model.

    The OneFormer model was proposed in OneFormer: One Transformer to Rule Universal Image Segmentation [1]
    by Jitesh Jain, Jiachen Li, MangTik Chiu, Ali Hassani, Nikita Orlov, Humphrey Shi. OneFormer is a universal
    image segmentation framework that can be trained on a single panoptic dataset to perform semantic, instance,
    and panoptic segmentation tasks.

    Whilst all official HuggingFace model weights by shi-labs [2] are supported, the current implementation only
    supports the OneFormer model for semantic segmentation.

    Instance and panoptic segmentation tasks are not supported yet.

    Usage Example:
    --------------
    >> from PIL import Image
    >> import numpy as np
    >> import torch
    >> from deepchem.data import NumpyDataset, ImageDataset
    >> from deepchem.models.torch_models import OneFormer
    >> from deepchem.metrics import Metric, mean_absolute_error, jaccaard_index

    >> # Prepare the dataset
    >> X = np.random.randint(0, 255, (3, 512, 512, 3))
    >> y = np.random.randint(0, 1, (3, 512, 512))
    >> dataset = ImageDataset(X, y)
    >> id2label = {0: "label-A", 1: "label-B"}

    >> # Create the model
    >> model = OneFormer(segmentation_task="semantic", model_path='shi-labs/oneformer_ade20k_swin_tiny',
    ...                  id2label=id2label, torch_dtype=torch.float16, batch_size=2)

    >> # Train the model
    >> avg_loss = model.fit(dataset, nb_epoch=3)

    >> # Predict the model
    >> preds = model.predict(dataset)

    >> # Evaluate the model
    >> mae_metric = Metric(mean_absolute_error)
    >> iou_metric = Metric(jaccaard_index)
    >> iou = iou_metric.compute_metric(dataset.y, preds)
    >> mae = mae_metric.compute_metric(np.array(dataset.y).flatten(), np.array(preds).flatten())

    References
    ----------
    .. [1] Jain, J., Li, J., Chiu, M., Hassani, A., Orlov, N., & Shi, H. (2022, November 10). OneFormer: One Transformer to rule universal image segmentation. arXiv.org. https://arxiv.org/abs/2211.06220
    .. [2] https://huggingface.co/shi-labs
    """

    def __init__(self,
                 segmentation_task: str = "semantic",
                 model_path: str = 'shi-labs/oneformer_ade20k_swin_tiny',
                 model_config: Optional[OneFormerConfig] = None,
                 model_processor: Optional[AutoProcessor] = None,
                 id2label: Dict = {
                     0: "unlabelled",
                     1: "labelled"
                 },
                 torch_dtype: torch.dtype = torch.float32,
                 **kwargs) -> None:
        """
        Parameters
        ----------
        segmentation_task: str
            The segmentation task to perform. The available tasks are
            - `semantic` - semantic segmentation (default)
            - `instance` - instance segmentation (not supported yet)
            - `panoptic` - panoptic segmentation (not supported yet)
        model_path: str
            Path to the OneFormer HuggingFace model; HF Model Hub or local
            - ex: 'shi-labs/oneformer_ade20k_swin_tiny'
        model_config: OneFormerConfig
            Optional configuration for the OneFormer model. If not provided, the configuration
            will be loaded from the model_path. If provided, the configuration will be used
            to initialize the model instead of the configuration from the model_path.
        model_processor: AutoProcessor
            Optional processor for the OneFormer model. If not provided, the processor
            will be loaded from the model_path. If provided, the processor will be used
            to initialize the model instead of the processor from the model_path.
        id2label: dict
            A dictionary mapping class indices to class labels.
        torch_dtype: torch.dtype
            The torch data type to use for the model. The supported data types are
            - `torch.float32` (default)
            - `torch.float16`

        Note
        ----
        If a custom model configuration and processor are provided, ensure that
        `model_processor.image_processor.num_text == model_config.num_queries - model_config.text_encoder_n_ctx` for
        contrastive learning to work correctly during training.
        """

        self.model_path = model_path
        self.segmentation_task = segmentation_task
        self.task = "universal_segmentation"
        self.id2label = id2label
        self.label2id = {v: k for k, v in self.id2label.items()}
        self.num_labels = len(self.id2label)
        self.torch_dtype = torch_dtype

        if model_config is None and model_path is not None:
            self.model_config = OneFormerConfig().from_pretrained(
                model_path, is_training=True, torch_dtype=torch_dtype)
        else:
            self.model_config = model_config

        if model_processor is None and model_path is not None:
            self.model_processor = AutoProcessor.from_pretrained(model_path)
        else:
            self.model_processor = model_processor

        if (self.model_path is None) and (self.model_config is None or
                                          self.model_processor is None):
            raise ValueError(
                "Please provide either a model path or a model configuration and processor."
            )

        self.model_config.id2label = self.id2label
        self.model_config.label2id = self.label2id

        self.model = AutoModelForUniversalSegmentation.from_config(
            self.model_config)

        self.model_processor.image_processor.num_text = self.model.config.num_queries - self.model.config.text_encoder_n_ctx
        assert self.model_processor.image_processor.num_text == self.model.config.num_queries - self.model.config.text_encoder_n_ctx

        super().__init__(
            model=self.model,
            task=self.task,
            tokenizer=None,  # type: ignore
            **kwargs)

    def _prepare_batch(self, batch: Tuple[Any, Any, Any]):
        """
        Preprocess and prepare the batch for the model.

        Parameters
        ----------
        batch: Tuple
            A tuple of the form (X, y, w) where X is the input data, y is the label and w is the weight.

        Returns
        -------
        processed_inputs: Dict
            A dictionary containing the processed inputs.
        y: torch.Tensor
            The label tensor.
        w: torch.Tensor
            The weight tensor.
        """
        X, y, w = batch
        if y is not None:
            X, y, w = X[0], y[0], w[0]
        else:
            X = X[0]

        images = []
        for x in X:
            images.append(Image.fromarray(x.astype(np.uint8)))

        self.image_size = images[0].size

        if y is not None:
            masks = y.astype(np.uint8)
        else:
            masks = None

        processed_inputs: Dict[str, torch.Tensor]
        processed_inputs = {}
        for idx, img in enumerate(images):
            if masks is not None:
                inputs = self.model_processor(
                    images=images[idx],
                    segmentation_maps=masks[idx],
                    task_inputs=[self.segmentation_task],
                    return_tensors="pt")
            else:
                inputs = self.model_processor(
                    images=images[idx],
                    task_inputs=[self.segmentation_task],
                    return_tensors="pt")
            # Process and append to the same dictionary
            for k, v in inputs.items():
                v = v.squeeze().to(self.device) if isinstance(
                    v, torch.Tensor) else v[0]

                # If this is the first time we're adding to processed_inputs, initialize it with the correct batch shape
                if k in processed_inputs:
                    processed_inputs[k] = torch.cat(
                        (processed_inputs[k], v.unsqueeze(0)), dim=0)
                else:
                    processed_inputs[k] = v.unsqueeze(0)  # Add batch dimension

        return processed_inputs, y, w

    def fit_generator(self,
                      generator: Iterable[Tuple[Any, Any, Any]],
                      max_checkpoints_to_keep: int = 5,
                      checkpoint_interval: int = 1000,
                      restore: bool = False,
                      variables: Optional[Union[List[torch.nn.Parameter],
                                                torch.nn.ParameterList]] = None,
                      loss: Optional[LossFn] = None,
                      callbacks: Union[Callable, List[Callable]] = [],
                      all_losses: Optional[List[float]] = None) -> float:
        """Train this model on data from a generator.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of torch.nn.Parameter
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step, **kwargs) that will be invoked
            after every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.

        Returns
        -------
        The average loss over the most recent checkpoint interval

        Note
        ----
        A HuggingFace model can return embeddings (last hidden state), attentions.
        Support must be added to return the embeddings to the user, so that it can
        be used for other downstream applications.
        """
        if not isinstance(callbacks, SequenceCollection):
            callbacks = [callbacks]
        self._ensure_built()
        self.model.train()
        self.model.model.is_training = True
        avg_loss = 0.0
        last_avg_loss = 0.0
        averaged_batches = 0
        if variables is None:
            optimizer = self._pytorch_optimizer
            lr_schedule = self._lr_schedule
        else:
            var_key = tuple(variables)
            if var_key in self._optimizer_for_vars:
                optimizer, lr_schedule = self._optimizer_for_vars[var_key]
            else:
                optimizer = self.optimizer._create_pytorch_optimizer(variables)
                if isinstance(self.optimizer.learning_rate,
                              LearningRateSchedule):
                    lr_schedule = self.optimizer.learning_rate._create_pytorch_schedule(
                        optimizer)
                else:
                    lr_schedule = None
                self._optimizer_for_vars[var_key] = (optimizer, lr_schedule)
        time1 = time.time()

        # Main training loop.
        for batch in generator:
            if restore:
                self.restore()
                restore = False

            inputs, labels, weights = self._prepare_batch(batch)
            inputs = {k: v.to(self.device) for k, v in inputs.items()}

            optimizer.zero_grad()
            outputs = self.model(**inputs)

            batch_loss = outputs.get("loss")
            batch_loss.backward()
            optimizer.step()
            if lr_schedule is not None:
                lr_schedule.step()
            self._global_step += 1
            current_step = self._global_step

            avg_loss += batch_loss

            # Report progress and write checkpoints.
            averaged_batches += 1
            should_log = (current_step % self.log_frequency == 0)
            if should_log:
                avg_loss = float(avg_loss) / averaged_batches
                logger.info('Ending global_step %d: Average loss %g' %
                            (current_step, avg_loss))
                if all_losses is not None:
                    all_losses.append(avg_loss)
                # Capture the last avg_loss in case of return since we're resetting to 0 now
                last_avg_loss = avg_loss
                avg_loss = 0.0
                averaged_batches = 0

            if checkpoint_interval > 0 and current_step % checkpoint_interval == checkpoint_interval - 1:
                self.save_checkpoint(max_checkpoints_to_keep)
            for c in callbacks:
                try:
                    # NOTE In DeepChem > 2.8.0, callback signature is updated to allow
                    # variable arguments.
                    c(self, current_step, iteration_loss=batch_loss)
                except TypeError:
                    # DeepChem <= 2.8.0, the callback should have this signature.
                    c(self, current_step)
            if self.tensorboard and should_log:
                self._log_scalar_to_tensorboard('loss', batch_loss,
                                                current_step)
            if (self.wandb_logger is not None) and should_log:
                all_data = dict({'train/loss': batch_loss})
                self.wandb_logger.log_data(all_data, step=current_step)

        # Report final results.
        if averaged_batches > 0:
            avg_loss = float(avg_loss) / averaged_batches
            logger.info('Ending global_step %d: Average loss %g' %
                        (current_step, avg_loss))
            if all_losses is not None:
                all_losses.append(avg_loss)
            last_avg_loss = avg_loss

        if checkpoint_interval > 0:
            self.save_checkpoint(max_checkpoints_to_keep)

        time2 = time.time()
        logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
        return last_avg_loss

    def _predict(self, generator: Iterable[Tuple[Any, Any, Any]],
                 transformers: List[Transformer], uncertainty: bool,
                 other_output_types: Optional[OneOrMany[str]]):
        """Predicts output for data provided by generator.

        This is the private implementation of prediction. Do not
        call it directly. Instead call one of the public prediction methods.
        Please use model.predict() instead.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        transformers: list of dc.trans.Transformers
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        uncertainty: bool
            specifies whether this is being called as part of estimating uncertainty.
            If True, it sets the training flag so that dropout will be enabled, and
            returns the values of the uncertainty outputs.
        other_output_types: list, optional
            Provides a list of other output_types (strings) to predict from model.

        Returns
        -------
            A NumPy array of the model produces a single output, or a list of arrays
            if it produces multiple outputs

        Note
        ----
        A HuggingFace model does not output uncertainity. The argument is here
        since it is also present in TorchModel. Similarly, other variables like
        other_output_types are also not used. Instead, a HuggingFace model outputs
        loss, logits, hidden state and attentions.
        """
        results: Optional[List[List[np.ndarray]]] = None
        variances: Optional[List[List[np.ndarray]]] = None
        if uncertainty and (other_output_types is not None):
            raise ValueError(
                'This model cannot compute uncertainties and other output types simultaneously. Please invoke one at a time.'
            )
        if uncertainty:
            if self._variance_outputs is None or len(
                    self._variance_outputs) == 0:
                raise ValueError('This model cannot compute uncertainties')
            if len(self._variance_outputs) != len(self._prediction_outputs):
                raise ValueError(
                    'The number of variances must exactly match the number of outputs'
                )
        if other_output_types:
            if self._other_outputs is None or len(self._other_outputs) == 0:
                raise ValueError(
                    'This model cannot compute other outputs since no other output_types were specified.'
                )
        self._ensure_built()
        self.model.eval()
        self.model.model.is_training = False

        for batch in generator:
            inputs, labels, weights = batch
            inputs, _, _ = self._prepare_batch((inputs, None, None))

            output_values = []
            # Invoke the model.
            with torch.no_grad():
                output = self.model(**inputs)
                output_values.append(output)

            # Post process the output
            for idx, i in enumerate(output_values):
                if self.segmentation_task == "semantic":
                    output_values[
                        idx] = self.model_processor.post_process_semantic_segmentation(
                            output_values[idx],
                            target_sizes=[self.image_size[::-1]])[0]

            if isinstance(output_values, torch.Tensor):
                output_values = [output_values]
            output_values = [t.detach().cpu().numpy() for t in output_values]

            # Apply tranformers and record results.
            if uncertainty:
                var = [output_values[i] for i in self._variance_outputs]
                if variances is None:
                    variances = [var]
                else:
                    for i, t in enumerate(var):
                        variances[i].append(t)
            access_values = []
            if other_output_types:
                access_values += self._other_outputs
            elif self._prediction_outputs is not None:
                access_values += self._prediction_outputs

            if len(access_values) > 0:
                output_values = [output_values[i] for i in access_values]

            if len(transformers) > 0:
                if len(output_values) > 1:
                    raise ValueError(
                        "predict() does not support Transformers for models with multiple outputs."
                    )
                elif len(output_values) == 1:
                    output_values = [
                        undo_transforms(output_values[0], transformers)
                    ]
            if results is None:
                results = [[] for i in range(len(output_values))]
            for i, t in enumerate(output_values):
                results[i].append(t)

        # Concatenate arrays to create the final results
        if results is not None:
            final_results = results
        if variances is not None:
            final_variances = variances

        if uncertainty and variances is not None:
            return zip(final_results, final_variances)

        if len(final_results) == 1:
            return final_results[0]
        else:
            return np.array(final_results)



import torch
import torch.nn as nn
from typing import Callable, Dict, List, Optional, Union
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.models.torch_models.layers import MultilayerPerceptron
import logging

logger = logging.getLogger(__name__)


class PINNModel(TorchModel):
    """
    This model is designed for solving linear partial differential equations (PDEs) using
    Physics-Informed Neural Networks (PINNs). It extends the TorchModel class, and its
    methods are similar to those in TorchModel, with additional functionality for handling
    physics-based constraints.

    Parameters
    ----------
    pde_fn : callable
        A function that defines the physics PDE residuals. Each PINN may have a unique
        strategy for calculating the physics losses, and this function specifies how
        the PINNModel computes the PDE residuals. The function should follow this format:

        >>> def heat_equation_residual(model, x):
        ...     x.requires_grad_(True)
        ...     u = model(x)
        ...     du_dx = torch.autograd.grad(u.sum(), x, create_graph=True, retain_graph=True)[0]
        ...     d2u_dx2 = torch.autograd.grad(du_dx.sum(), x, create_graph=True, retain_graph=True)[0]
        ...     return d2u_dx2

        Here, `model` is the neural network being trained, and `x` is the input.

    boundary_data : dict
        A dictionary containing the boundary condition data. The PINNModel supports three boundary data types: Dirichlet, Neumann and Robin.
        The following format must be followed:

        >>> boundary_data = {
        ...     'dirichlet': {
        ...         'points': torch.tensor([[0.0], [1.0]], dtype=torch.float32),
        ...         'values': torch.tensor([[0.0], [1.0]], dtype=torch.float32)
        ...     }
        ... }

        - `points`: Tensor of input points where boundary conditions are defined.
        - `values`: Tensor of target values at the boundary points.

    loss_fn : callable
        A custom loss function that combines data, physics, and boundary losses.
        An example is shown below:

        >>> def custom_loss(outputs, labels, weights=None):
        ...     outputs = outputs[0]
        ...     labels = labels[0]
        ...     data_loss = torch.mean(torch.square(outputs - labels))
        ...     pde_residuals = heat_equation_residual(model, labels)
        ...     pde_loss = torch.mean(torch.abs(pde_residuals))
        ...     boundary_loss = 0.0
        ...     for _, value in boundary_data.items():
        ...         if isinstance(value, dict):
        ...             points = value.get('points')
        ...             values = value.get('values')
        ...             if points is not None and values is not None:
        ...                 pred = model(points)
        ...                 boundary_loss += torch.mean(torch.square(pred - values))
        ...     return data_loss + pde_loss + 10 * boundary_loss

    Usage Example
    -------------
    Here's an example of using PINNModel to solve the 1D steady-state heat equation:

    >>> import torch
    >>> import deepchem as dc
    >>> class HeatNet(torch.nn.Module):
    ...     def __init__(self):
    ...         super(HeatNet, self).__init__()
    ...         self.net = torch.nn.Sequential(
    ...             torch.nn.Linear(1, 64),
    ...             torch.nn.Tanh(),
    ...             torch.nn.Linear(64, 64),
    ...             torch.nn.Tanh(),
    ...             torch.nn.Linear(64, 1)
    ...         )
    ...     def forward(self, x):
    ...         if not isinstance(x, torch.Tensor):
    ...             x = torch.tensor(x, dtype=torch.float32)
    ...         return self.net(x)
    >>> def heat_equation_residual(u, x):
    ...     x.requires_grad_(True)
    ...     du_dx = torch.autograd.grad(u.sum(), x, create_graph=True, retain_graph=True)[0]
    ...     d2u_dx2 = torch.autograd.grad(du_dx.sum(), x, create_graph=True, retain_graph=True)[0]
    ...     return du_dx - d2u_dx2  # Let alpha be 1.0
    >>> x_interior = torch.linspace(0, 1, 2000)[1:-1].reshape(-1, 1)
    >>> x_boundary = torch.tensor([[0.0], [1.0]])
    >>> x = torch.cat([x_interior, x_boundary], dim=0)
    >>> y = x.clone()
    >>> dataset = dc.data.NumpyDataset(X=x.numpy(), y=y.numpy())
    >>> boundary_data = {
    ...     'dirichlet': {
    ...         'points': torch.tensor([[0.0], [1.0]], dtype=torch.float32),
    ...         'values': torch.tensor([[0.0], [1.0]], dtype=torch.float32)
    ...     }
    ... }
    >>> model = HeatNet()
    >>> pinn = PINNModel(
    ...     model=model,
    ...     pde_fn=heat_equation_residual,
    ...     boundary_data=boundary_data,
    ... )
    >>> loss = pinn.fit(dataset, nb_epoch=100)

    References
    ----------
    [1] Raissi et al. "Physics-informed neural networks: A deep learning framework for solving
        forward and inverse problems involving nonlinear partial differential equations."
        Journal of Computational Physics, https://doi.org/10.1016/j.jcp.2018.10.045

    [2] Raissi et al. "Physics-Informed Deep Learning (Part I): Data-driven Solutions of Nonlinear
        Partial Differential Equations." arXiv preprint arXiv:1711.10561

    Notes
    -----
    - This class requires PyTorch to be installed.
    - Users can use the default neural network provided by the class or pass a custom model.
    """

    def __init__(self,
                 model: Optional[nn.Module] = None,
                 in_features: Optional[int] = None,
                 loss_fn: Optional[Callable] = None,
                 pde_fn: Union[List,
                               Callable] = [lambda u, x: torch.zeros_like(u)],
                 pde_weights: Union[List, float] = [1.0],
                 boundary_data: Dict = {},
                 data_weight: float = 1.0,
                 physics_weight: float = 1.0,
                 eval_fn: Optional[Callable] = None,
                 mode: str = 'regression',
                 **kwargs) -> None:
        """
        Initialize PINNModel.

        Parameters
        ----------
        model : nn.Module, optional
            PyTorch neural network model for training. If not provided, a default neural network is used for regression.
        in_features : int, optional
            Number of input features for the default model. Ignored if a custom model is provided.
        loss_fn : Callable
            Loss function for the data-driven part of the training.
        pde_fn : Callable or List[Callable]
            Function(s) that compute the PDE residuals. Should take model predictions and input
            coordinates as arguments and return the PDE residuals.
        pde_weights : float or List[float]
            Weights for each PDE when there are multiple. If a single value is provided, it is applied to all PDE terms.
        boundary_data : Dict
            Dictionary containing boundary condition data.
        data_weight : float, optional, default=1.0
            Weight for the data-driven loss term in the total loss computation.
        physics_weight : float, optional, default=1.0
            Weight for the physics-informed loss term in the total loss computation.
        eval_fn : Callable, optional
            Custom function for model evaluation during inference. If not provided, a default
            evaluation function is used.
        **kwargs :
            Additional arguments passed to the parent `TorchModel` class.
        """

        if model is None:
            # Regression by default
            if in_features is not None:
                model = MultilayerPerceptron(d_input=in_features,
                                             d_hidden=(100, 100, 100, 100, 100),
                                             d_output=1,
                                             activation_fn='tanh')
            else:
                model = MultilayerPerceptron(d_input=1,
                                             d_hidden=(100, 100, 100, 100, 100),
                                             d_output=1,
                                             activation_fn='tanh')

        if not isinstance(pde_fn, list):
            pde_fn = [pde_fn]
        if not isinstance(pde_weights, list):
            pde_weights = [pde_weights] * len(pde_fn)

        self.mode = mode
        self.loss_fn = loss_fn or self._loss_fn
        if loss_fn is None:
            self.data_loss_fn = nn.MSELoss(
            ) if self.mode == 'regression' else nn.CrossEntropyLoss()
        self.pde_fn = [pde_fn] if not isinstance(pde_fn, list) else pde_fn
        self.boundary_data = boundary_data
        self.eval_fn = eval_fn or self._default_eval_fn
        self.data_weight = data_weight
        self.physics_weight = physics_weight
        self.pde_weights = ([pde_weights] * len(self.pde_fn) if
                            not isinstance(pde_weights, list) else pde_weights)
        super(PINNModel, self).__init__(model=model,
                                        loss=self.loss_fn,
                                        **kwargs)

    def _compute_pde_loss(self, x: torch.Tensor) -> torch.Tensor:
        """
        Compute the physics-informed loss using PDE residuals.

        The physics loss ensures that the model satisfies the given PDEs by minimizing
        the residuals computed for the input coordinates.

        Parameters
        ----------
        x : torch.Tensor
            Input coordinates where PDE residuals are computed. Gradients are enabled
            on `x` to allow for backpropagation through the PDE computations.

        Returns
        -------
        torch.Tensor
            The computed physics-informed loss.
        """

        x.requires_grad_(True)
        u_pred = self.model(x)
        residual_loss = torch.tensor(0.0)
        for pde, weight in zip(self.pde_fn, self.pde_weights):
            residuals = pde(u_pred, x)
            residual_loss += weight * torch.mean(torch.square(residuals))
        return residual_loss

    def _compute_boundary_loss(self) -> torch.Tensor:
        """
        Compute the loss at boundary points for Dirichlet, Neumann, and Robin conditions.

        This loss ensures that the model satisfies the specified boundary conditions:
            - Dirichlet: Model outputs match the target values at boundary points.
            - Neumann: Gradients of model outputs match target values at boundary points.
            - Robin: A linear combination of model outputs and their gradients matches target values.

        Parameters
        ----------
        None

        Returns
        -------
        torch.Tensor
            The computed boundary loss.
        """

        if not self.boundary_data:
            return torch.tensor(0.0)

        boundary_loss = torch.tensor(0.0)
        for key, value in self.boundary_data.items():
            if isinstance(value, dict):
                points = value.get('points')
                values = value.get('values')
                alpha = value.get('alpha', 1.0)
                beta = value.get('beta', 1.0)
                if points is not None and values is not None:
                    pred = self.model(points)
                    if key.lower() == 'dirichlet':
                        boundary_loss += torch.mean(torch.square(pred - values))
                    elif key.lower() == 'neumann':
                        pred_grad = torch.autograd.grad(
                            pred,
                            points,
                            grad_outputs=torch.ones_like(pred),
                            create_graph=True)[0]
                        boundary_loss += torch.mean(
                            torch.square(pred_grad - values))
                    elif key.lower() == 'robin':
                        pred_grad = torch.autograd.grad(
                            pred,
                            points,
                            grad_outputs=torch.ones_like(pred),
                            create_graph=True)[0]
                        robin_loss = alpha * pred + beta * pred_grad - values
                        boundary_loss += torch.mean(torch.square(robin_loss))
                    else:
                        logger.warning(
                            f"Unknown boundary condition type: {key}. Skipping boundary value."
                        )
                        continue

        return boundary_loss

    def _default_eval_fn(self, x: torch.Tensor) -> torch.Tensor:
        """Default evaluation function for inference."""
        with torch.no_grad():
            return self.model(x)

    def _prepare_batch(self, batch):
        """
        Overrides the parent class's method to save the input tensors for later use.

        Parameters
        ----------
        batch : dict
            A dictionary containing the batch data, which includes inputs,
            labels, and optionally weights for the loss calculation.

        Returns
        -------
        tuple
            A tuple containing:
            - input_tensors (torch.Tensor): The input data for the model.
            - label_tensors (torch.Tensor): The ground truth values corresponding to the inputs.
            - weight_tensors (torch.Tensor or None): Optional weights for the loss terms, if provided.
        """
        input_tensors, label_tensors, weight_tensors = super()._prepare_batch(
            batch)
        self._inputs = input_tensors
        return input_tensors, label_tensors, weight_tensors

    def _loss_fn(self, outputs: list, labels: list,
                 weights: list) -> torch.Tensor:
        """
        Custom loss function that combines data, physics, and boundary losses.

        The total loss is calculated as:
            Total loss = data_weight * data_loss + physics_weight * physics_loss + boundary_loss

        Parameters
        ----------
        outputs : list
            The model's predictions.
        labels : list
            The ground truth values corresponding to the predictions.
        weights : list
            Optional weights for the loss terms.

        Returns
        -------
        torch.Tensor
            The computed total loss.
        """

        outputs = outputs[0]
        labels = labels[0]
        inputs = self._inputs[0]

        data_loss = self.data_loss_fn(outputs, labels)
        physics_loss = self._compute_pde_loss(inputs)
        boundary_loss = self._compute_boundary_loss()

        total_loss = self.data_weight * data_loss + self.physics_weight * physics_loss + boundary_loss

        return total_loss

    def predict(self, dataset):
        """
        Makes predictions on dataset using the evaluation function.

        Parameters
        ----------
        dataset: Dataset
            Dataset to make predictions on

        Returns
        -------
        Predicted values
        """
        return self.eval_fn(dataset)



import re
import warnings
try:
    import pylibxc
except (ImportError, ModuleNotFoundError) as e:
    warnings.warn(f"Failed to import pylibxc. Might not be able to use xc. {e}")
from deepchem.utils.dft_utils import BaseXC, LibXCLDA, LibXCGGA, LibXCMGGA


def get_libxc(name: str) -> BaseXC:
    """
    Get the XC object of the libxc based on its libxc's name.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> xc = get_libxc("gga_c_pbe")
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> densinfo = ValGrad(value=rho_u, grad=grad_u)
    >>> # get the exchange-correlation potential
    >>> potinfo = xc.get_vxc(densinfo)
    >>> potinfo.value.shape
    torch.Size([2])

    Parameters
    ----------
    name: str
        The full libxc name, e.g. "lda_c_pw"

    Returns
    -------
    BaseXC
        XC object that wraps the xc requested

    """
    obj = pylibxc.LibXCFunctional(name, "unpolarized")
    family = obj.get_family()
    del obj
    if family == 1:  # LDA
        return LibXCLDA(name)
    elif family == 2:  # GGA
        return LibXCGGA(name)
    elif family == 4:  # MGGA
        return LibXCMGGA(name)
    else:
        raise NotImplementedError(
            "LibXC wrapper for family %d has not been implemented" % family)


def get_xc(xcstr: str) -> BaseXC:
    """
    Returns the XC object based on the expression in xcstr.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import ValGrad
    >>> xc = get_xc("lda_x + gga_c_pbe")
    >>> n = 2
    >>> rho_u = torch.rand((n,), dtype=torch.float64).requires_grad_()
    >>> grad_u = torch.rand((3, n), dtype=torch.float64).requires_grad_()
    >>> densinfo = ValGrad(value=rho_u, grad=grad_u)
    >>> # get the exchange-correlation potential
    >>> potinfo = xc.get_vxc(densinfo)
    >>> potinfo.value.shape
    torch.Size([2])

    Parameters
    ----------
    xcstr: str
        The expression of the xc string, e.g. ``"lda_x + gga_c_pbe"`` where the
        variable name will be replaced by the LibXC object

    Returns
    -------
    BaseXC
        XC object based on the given expression

    """
    # wrap the name of xc with "get_libxc"
    pattern = r"([a-zA-Z_$][a-zA-Z_$0-9]*)"
    new_xcstr = re.sub(pattern, r'get_libxc("\1")', xcstr)

    # evaluate the expression and return the xc
    glob = {"get_libxc": get_libxc}
    return eval(new_xcstr, glob)



"""
Sample supports from datasets.
"""
import logging
import time
import numpy as np
from deepchem.data import NumpyDataset

logger = logging.getLogger(__name__)


def remove_dead_examples(dataset):
    """Removes compounds with no weight.

    Parameters
    ----------
    dataset: dc.data.Dataset
        Source dataset.
    """
    w = dataset.w
    nonzero_inds = np.nonzero(np.sum(w, axis=1))

    # Remove support indices
    X = dataset.X[nonzero_inds]
    y = dataset.y[nonzero_inds]
    w = dataset.w[nonzero_inds]
    ids = dataset.ids[nonzero_inds]

    return NumpyDataset(X, y, w, ids)


def dataset_difference(dataset, remove):
    """Removes the compounds in remove from dataset.

    Parameters
    ----------
    dataset: dc.data.Dataset
        Source dataset.
    remove: dc.data.Dataset
        Dataset whose overlap will be removed.
    """
    remove_ids = set(remove.ids)
    keep_inds = [
        ind for ind in range(len(dataset)) if dataset.ids[ind] not in remove_ids
    ]

    # Remove support indices
    X = dataset.X[keep_inds]
    y = dataset.y[keep_inds]
    w = dataset.w[keep_inds]
    ids = dataset.ids[keep_inds]

    return NumpyDataset(X, y, w, ids)


def get_task_dataset_minus_support(dataset, support, task):
    """Gets data for specified task, minus support points.

    Useful for evaluating model performance once trained (so that
    test compounds can be ensured distinct from support.)

    Parameters
    ----------
    dataset: dc.data.Dataset
        Source dataset.
    support: dc.data.Dataset
        The support dataset
    task: int
        Task number of task to select.
    """
    support_ids = set(support.ids)
    non_support_inds = [
        ind for ind in range(len(dataset))
        if dataset.ids[ind] not in support_ids
    ]

    # Remove support indices
    X = dataset.X[non_support_inds]
    y = dataset.y[non_support_inds]
    w = dataset.w[non_support_inds]
    ids = dataset.ids[non_support_inds]

    # Get task specific entries
    w_task = w[:, task]
    X_task = X[w_task != 0]
    y_task = np.expand_dims(y[w_task != 0, task], 1)
    ids_task = ids[w_task != 0]
    # Now just get weights for this task
    w_task = np.expand_dims(w[w_task != 0, task], 1)

    return NumpyDataset(X_task, y_task, w_task, ids_task)


def get_task_dataset(dataset, task):
    """Selects out entries for a particular task."""
    X, y, w, ids = dataset.X, dataset.y, dataset.w, dataset.ids
    # Get task specific entries
    w_task = w[:, task]
    X_task = X[w_task != 0]
    y_task = np.expand_dims(y[w_task != 0, task], 1)
    ids_task = ids[w_task != 0]
    # Now just get weights for this task
    w_task = np.expand_dims(w[w_task != 0, task], 1)

    return NumpyDataset(X_task, y_task, w_task, ids_task)


def get_task_test(dataset, n_episodes, n_test, task, log_every_n=50):
    """Gets test set from specified task.

    Parameters
    ----------
    dataset: dc.data.Dataset
        Dataset from which to sample.
    n_episodes: int
        Number of episodes to sample test sets for.
    n_test: int
        Number of compounds per test set.
    log_every_n: int, optional
        Prints every log_every_n supports sampled.
    """
    w_task = dataset.w[:, task]
    X_task = dataset.X[w_task != 0]
    y_task = dataset.y[w_task != 0]
    ids_task = dataset.ids[w_task != 0]
    # Now just get weights for this task
    w_task = dataset.w[w_task != 0]

    n_samples = len(X_task)

    ids = np.random.choice(np.arange(n_samples), (n_episodes, n_test))

    tests = []
    for episode in range(n_episodes):
        if episode % log_every_n == 0:
            logger.info("Sampling test %d" % episode)
        inds = ids[episode]
        X_batch = X_task[inds]
        y_batch = np.squeeze(y_task[inds, task])
        w_batch = np.squeeze(w_task[inds, task])
        ids_batch = ids_task[inds]
        tests.append(NumpyDataset(X_batch, y_batch, w_batch, ids_batch))
    return tests


def get_single_task_test(dataset, batch_size, task, replace=True):
    """Gets test set from specified task.

    Samples random subset of size batch_size from specified task of dataset.
    Ensures that sampled points have measurements for this task.
    """
    w_task = dataset.w[:, task]
    X_task = dataset.X[w_task != 0]
    y_task = dataset.y[w_task != 0]
    ids_task = dataset.ids[w_task != 0]
    # Now just get weights for this task
    w_task = dataset.w[w_task != 0]

    inds = np.random.choice(np.arange(len(X_task)), batch_size, replace=replace)
    X_batch = X_task[inds]
    y_batch = np.squeeze(y_task[inds, task])
    w_batch = np.squeeze(w_task[inds, task])
    ids_batch = ids_task[inds]
    return NumpyDataset(X_batch, y_batch, w_batch, ids_batch)


def get_single_task_support(dataset, n_pos, n_neg, task, replace=True):
    """Generates one support set purely for specified task.

    Parameters
    ----------
    datasets: dc.data.Dataset
        Dataset from which supports are sampled.
    n_pos: int
        Number of positive samples in support.
    n_neg: int
        Number of negative samples in support.
    task: int
        Index of current task.
    replace: bool, optional
        Whether or not to use replacement when sampling supports.

    Returns
    -------
    list
        List of NumpyDatasets, each of which is a support set.
    """
    return get_task_support(dataset, 1, n_pos, n_neg, task)[0]


def get_task_support(dataset, n_episodes, n_pos, n_neg, task, log_every_n=50):
    """Generates one support set purely for specified task.

    Parameters
    ----------
    datasets: dc.data.Dataset
        Dataset from which supports are sampled.
    n_episodes: int
        Number of episodes for which supports have to be sampled from this task.
    n_pos: int
        Number of positive samples in support.
    n_neg: int
        Number of negative samples in support.
    task: int
        Index of current task.
    log_every_n: int, optional
        Prints every log_every_n supports sampled.

    Returns
    -------
    list
        List of NumpyDatasets, each of which is a support set.
    """
    y_task = dataset.y[:, task]
    w_task = dataset.w[:, task]

    # Split data into pos and neg lists.
    pos_mols = np.where(np.logical_and(y_task == 1, w_task != 0))[0]
    neg_mols = np.where(np.logical_and(y_task == 0, w_task != 0))[0]

    supports = []
    for episode in range(n_episodes):
        if episode % log_every_n == 0:
            logger.info("Sampling support %d" % episode)
        # No replacement allowed for supports
        pos_ids = np.random.choice(len(pos_mols), (n_pos,), replace=False)
        neg_ids = np.random.choice(len(neg_mols), (n_neg,), replace=False)
        pos_inds, neg_inds = pos_mols[pos_ids], neg_mols[neg_ids]
        # Handle one-d vs. non one-d feature matrices
        one_dimensional_features = (len(dataset.X.shape) == 1)
        if not one_dimensional_features:
            X = np.vstack([dataset.X[pos_inds], dataset.X[neg_inds]])
        else:
            X = np.concatenate([dataset.X[pos_inds], dataset.X[neg_inds]])
        y = np.expand_dims(
            np.concatenate(
                [dataset.y[pos_inds, task], dataset.y[neg_inds, task]]), 1)
        w = np.expand_dims(
            np.concatenate(
                [dataset.w[pos_inds, task], dataset.w[neg_inds, task]]), 1)
        ids = np.concatenate([dataset.ids[pos_inds], dataset.ids[neg_inds]])
        supports.append(NumpyDataset(X, y, w, ids))
    return supports


class EpisodeGenerator(object):
    """Generates (support, test) pairs for episodic training.

    Precomputes all (support, test) pairs at construction. Allows to reduce
    overhead from computation.
    """

    def __init__(self, dataset, n_pos, n_neg, n_test, n_episodes_per_task):
        """
        Parameters
        ----------
        dataset: dc.data.Dataset
            Holds dataset from which support sets will be sampled.
        n_pos: int
            Number of positive samples
        n_neg: int
            Number of negative samples.
        n_test: int
            Number of samples in test set.
        n_episodes_per_task: int
            Number of (support, task) pairs to sample per task.
        replace: bool
            Whether to use sampling with or without replacement.
        """
        time_start = time.time()
        self.tasks = range(len(dataset.get_task_names()))
        self.n_tasks = len(self.tasks)
        self.n_episodes_per_task = n_episodes_per_task
        self.dataset = dataset
        self.n_pos = n_pos
        self.n_neg = n_neg
        self.task_episodes = {}

        for task in range(self.n_tasks):
            task_supports = get_task_support(self.dataset, n_episodes_per_task,
                                             n_pos, n_neg, task)
            task_tests = get_task_test(self.dataset, n_episodes_per_task,
                                       n_test, task)
            self.task_episodes[task] = (task_supports, task_tests)

        # Init the iterator
        self.perm_tasks = np.random.permutation(self.tasks)
        # Set initial iterator state
        self.task_num = 0
        self.trial_num = 0
        time_end = time.time()
        logger.info("Constructing EpisodeGenerator took %s seconds" %
                    str(time_end - time_start))

    def __iter__(self):
        return self

    def next(self):
        """Sample next (support, test) pair.

        Return from internal storage.
        """
        if self.trial_num == self.n_episodes_per_task:
            raise StopIteration
        else:
            task = self.perm_tasks[self.task_num]  # Get id from permutation
            # support = self.supports[task][self.trial_num]
            task_supports, task_tests = self.task_episodes[task]
            support, test = (task_supports[self.trial_num],
                             task_tests[self.trial_num])
            # Increment and update logic
            self.task_num += 1
            if self.task_num == self.n_tasks:
                self.task_num = 0  # Reset
                self.perm_tasks = np.random.permutation(
                    self.tasks)  # Permute again
                self.trial_num += 1  # Upgrade trial index

            return (task, support, test)

    __next__ = next  # Python 3.X compatibility


class SupportGenerator(object):
    """Generate support sets from a dataset.

    Iterates over tasks and trials. For each trial, picks one support from
    each task, and returns in a randomized order
    """

    def __init__(self, dataset, n_pos, n_neg, n_trials):
        """
        Parameters
        ----------
        dataset: dc.data.Dataset
            Holds dataset from which support sets will be sampled.
        n_pos: int
            Number of positive samples
        n_neg: int
            Number of negative samples.
        n_trials: int
            Number of passes over dataset to make. In total, n_tasks*n_trials
            support sets will be sampled by algorithm.
        """

        self.tasks = range(len(dataset.get_task_names()))
        self.n_tasks = len(self.tasks)
        self.n_trials = n_trials
        self.dataset = dataset
        self.n_pos = n_pos
        self.n_neg = n_neg

        # Init the iterator
        self.perm_tasks = np.random.permutation(self.tasks)
        # Set initial iterator state
        self.task_num = 0
        self.trial_num = 0

    def __iter__(self):
        return self

    def next(self):
        """Sample next support.

        Supports are sampled from the tasks in a random order. Each support is
        drawn entirely from within one task.
        """
        if self.trial_num == self.n_trials:
            raise StopIteration
        else:
            task = self.perm_tasks[self.task_num]  # Get id from permutation
            # support = self.supports[task][self.trial_num]
            support = get_single_task_support(self.dataset,
                                              n_pos=self.n_pos,
                                              n_neg=self.n_neg,
                                              task=task,
                                              replace=False)
            # Increment and update logic
            self.task_num += 1
            if self.task_num == self.n_tasks:
                self.task_num = 0  # Reset
                self.perm_tasks = np.random.permutation(
                    self.tasks)  # Permute again
                self.trial_num += 1  # Upgrade trial index

            return (task, support)

    __next__ = next  # Python 3.X compatibility



import logging
import time

from deepchem.utils.rdkit_utils import MoleculeLoadException, load_molecule, compute_ecfp_features
from deepchem.utils.geometry_utils import rotate_molecules, compute_pairwise_distances, compute_centroid, subtract_centroid
from deepchem.utils.hash_utils import hash_ecfp, hash_ecfp_pair, hash_sybyl, vectorize
from deepchem.utils.noncovalent_utils import compute_hydrogen_bonds, compute_salt_bridges, compute_binding_pocket_cation_pi
from deepchem.utils.voxel_utils import convert_atom_to_voxel, convert_atom_pair_to_voxel, voxelize, voxelize_pi_stack

from deepchem.feat.complex_featurizers.contact_fingerprints import featurize_contacts_ecfp, featurize_binding_pocket_sybyl
from deepchem.feat.complex_featurizers.splif_fingerprints import featurize_splif
from deepchem.feat.complex_featurizers.grid_featurizers import compute_charge_dictionary

import numpy as np
from deepchem.feat import ComplexFeaturizer

logger = logging.getLogger(__name__)


class RdkitGridFeaturizer(ComplexFeaturizer):
    """Featurizes protein-ligand complex using flat features or a 3D grid (in which
    each voxel is described with a vector of features).
    """

    def __init__(self,
                 nb_rotations=0,
                 feature_types=None,
                 ecfp_degree=2,
                 ecfp_power=3,
                 splif_power=3,
                 box_width=16.0,
                 voxel_width=1.0,
                 flatten=False,
                 verbose=True,
                 sanitize=False,
                 **kwargs):
        """
        Parameters
        ----------
        nb_rotations: int, optional (default 0)
            Number of additional random rotations of a complex to generate.
        feature_types: list, optional (default ['ecfp'])
            Types of features to calculate. Available types are
                flat features -> 'ecfp_ligand', 'ecfp_hashed', 'splif_hashed', 'hbond_count'
                voxel features -> 'ecfp', 'splif', 'sybyl', 'salt_bridge', 'charge', 'hbond', 'pi_stack, 'cation_pi'
            There are also 3 predefined sets of features
                'flat_combined', 'voxel_combined', and 'all_combined'.
            Calculated features are concatenated and their order is preserved
            (features in predefined sets are in alphabetical order).
        ecfp_degree: int, optional (default 2)
            ECFP radius.
        ecfp_power: int, optional (default 3)
            Number of bits to store ECFP features (resulting vector will be
            2^ecfp_power long)
        splif_power: int, optional (default 3)
            Number of bits to store SPLIF features (resulting vector will be
            2^splif_power long)
        box_width: float, optional (default 16.0)
            Size of a box in which voxel features are calculated. Box is centered on a
            ligand centroid.
        voxel_width: float, optional (default 1.0)
            Size of a 3D voxel in a grid.
        flatten: bool, optional (defaul False)
            Indicate whether calculated features should be flattened. Output is always
            flattened if flat features are specified in feature_types.
        verbose: bool, optional (defaul True)
            Verbolity for logging
        sanitize: bool, optional (defaul False)
            If set to True molecules will be sanitized. Note that calculating some
            features (e.g. aromatic interactions) require sanitized molecules.
        **kwargs: dict, optional
            Keyword arguments can be usaed to specify custom cutoffs and bins (see
            default values below).

        Default cutoffs and bins
        ------------------------
        hbond_dist_bins: [(2.2, 2.5), (2.5, 3.2), (3.2, 4.0)]
        hbond_angle_cutoffs: [5, 50, 90]
        splif_contact_bins: [(0, 2.0), (2.0, 3.0), (3.0, 4.5)]
        ecfp_cutoff: 4.5
        sybyl_cutoff: 7.0
        salt_bridges_cutoff: 5.0
        pi_stack_dist_cutoff: 4.4
        pi_stack_angle_cutoff: 30.0
        cation_pi_dist_cutoff: 6.5
        cation_pi_angle_cutoff: 30.0
        """

        # check if user tries to set removed arguments
        deprecated_args = [
            'box_x', 'box_y', 'box_z', 'save_intermediates',
            'voxelize_features', 'parallel', 'voxel_feature_types'
        ]

        # list of features that require sanitized molecules
        require_sanitized = ['pi_stack', 'cation_pi', 'ecfp_ligand']

        # not implemented featurization types
        not_implemented = ['sybyl']

        for arg in deprecated_args:
            if arg in kwargs and verbose:
                logger.warning(
                    '%s argument was removed and it is ignored,'
                    ' using it will result in error in version 1.4' % arg,
                    DeprecationWarning)

        self.verbose = verbose
        self.sanitize = sanitize
        self.flatten = flatten

        self.ecfp_degree = ecfp_degree
        self.ecfp_power = ecfp_power
        self.splif_power = splif_power

        self.nb_rotations = nb_rotations

        # default values
        self.cutoffs = {
            'hbond_dist_bins': [(2.2, 2.5), (2.5, 3.2), (3.2, 4.0)],
            'hbond_angle_cutoffs': [5, 50, 90],
            'splif_contact_bins': [(0, 2.0), (2.0, 3.0), (3.0, 4.5)],
            'ecfp_cutoff': 4.5,
            'sybyl_cutoff': 7.0,
            'salt_bridges_cutoff': 5.0,
            'pi_stack_dist_cutoff': 4.4,
            'pi_stack_angle_cutoff': 30.0,
            'cation_pi_dist_cutoff': 6.5,
            'cation_pi_angle_cutoff': 30.0,
        }

        # update with cutoffs specified by the user
        for arg, value in kwargs.items():
            if arg in self.cutoffs:
                self.cutoffs[arg] = value

        self.box_width = float(box_width)
        self.voxel_width = float(voxel_width)
        self.voxels_per_edge = int(self.box_width / self.voxel_width)

        self.sybyl_types = [
            "C3", "C2", "C1", "Cac", "Car", "N3", "N3+", "Npl", "N2", "N1",
            "Ng+", "Nox", "Nar", "Ntr", "Nam", "Npl3", "N4", "O3", "O-", "O2",
            "O.co2", "O.spc", "O.t3p", "S3", "S3+", "S2", "So2", "Sox"
            "Sac"
            "SO", "P3", "P", "P3+", "F", "Cl", "Br", "I"
        ]

        self.FLAT_FEATURES = [
            'ecfp_ligand', 'ecfp_hashed', 'splif_hashed', 'hbond_count'
        ]
        self.VOXEL_FEATURES = [
            'ecfp', 'splif', 'sybyl', 'salt_bridge', 'charge', 'hbond',
            'pi_stack', 'cation_pi'
        ]

        if feature_types is None:
            feature_types = ['ecfp']

        # each entry is a tuple (is_flat, feature_name)
        self.feature_types = []

        # list of features that cannot be calculated with specified parameters
        # this list is used to define <flat/voxel/all>_combined subset
        ignored_features = []
        if self.sanitize is False:
            ignored_features += require_sanitized
        ignored_features += not_implemented

        # parse provided feature types
        for feature_type in feature_types:
            if self.sanitize is False and feature_type in require_sanitized:
                if self.verbose:
                    logger.warning(
                        'sanitize is set to False, %s feature will be ignored' %
                        feature_type)
                continue
            if feature_type in not_implemented:
                if self.verbose:
                    logger.warning(
                        '%s feature is not implemented yet and will be ignored'
                        % feature_type)
                continue

            if feature_type in self.FLAT_FEATURES:
                self.feature_types.append((True, feature_type))
                if self.flatten is False:
                    if self.verbose:
                        logger.warning(
                            '%s feature is used, output will be flattened' %
                            feature_type)
                    self.flatten = True

            elif feature_type in self.VOXEL_FEATURES:
                self.feature_types.append((False, feature_type))

            elif feature_type == 'flat_combined':
                self.feature_types += [(True, ftype)
                                       for ftype in sorted(self.FLAT_FEATURES)
                                       if ftype not in ignored_features]
                if self.flatten is False:
                    if self.verbose:
                        logger.warning(
                            'Flat features are used, output will be flattened')
                    self.flatten = True

            elif feature_type == 'voxel_combined':
                self.feature_types += [(False, ftype)
                                       for ftype in sorted(self.VOXEL_FEATURES)
                                       if ftype not in ignored_features]
            elif feature_type == 'all_combined':
                self.feature_types += [(True, ftype)
                                       for ftype in sorted(self.FLAT_FEATURES)
                                       if ftype not in ignored_features]
                self.feature_types += [(False, ftype)
                                       for ftype in sorted(self.VOXEL_FEATURES)
                                       if ftype not in ignored_features]
                if self.flatten is False:
                    if self.verbose:
                        logger.warning(
                            'Flat feature are used, output will be flattened')
                    self.flatten = True
            elif self.verbose:
                logger.warning('Ignoring unknown feature %s' % feature_type)

    def _compute_feature(self, feature_name, prot_xyz, prot_rdk, lig_xyz,
                         lig_rdk, distances):
        if feature_name == 'ecfp_ligand':
            return [
                compute_ecfp_features(lig_rdk, self.ecfp_degree,
                                      self.ecfp_power)
            ]
        if feature_name == 'ecfp_hashed':
            return [
                vectorize(hash_ecfp,
                          feature_dict=ecfp_dict,
                          size=2**self.ecfp_power)
                for ecfp_dict in featurize_contacts_ecfp(
                    (prot_xyz, prot_rdk), (lig_xyz, lig_rdk),
                    distances,
                    cutoff=self.cutoffs['ecfp_cutoff'],
                    ecfp_degree=self.ecfp_degree)
            ]
        if feature_name == 'splif_hashed':
            return [
                vectorize(hash_ecfp_pair,
                          feature_dict=splif_dict,
                          size=2**self.splif_power)
                for splif_dict in featurize_splif((prot_xyz, prot_rdk), (
                    lig_xyz, lig_rdk), self.cutoffs['splif_contact_bins'],
                                                  distances, self.ecfp_degree)
            ]
        if feature_name == 'hbond_count':
            return [
                vectorize(hash_ecfp_pair, feature_list=hbond_list, size=2**0)
                for hbond_list in
                compute_hydrogen_bonds((prot_xyz, prot_rdk), (
                    lig_xyz,
                    lig_rdk), distances, self.cutoffs['hbond_dist_bins'],
                                       self.cutoffs['hbond_angle_cutoffs'])
            ]
        if feature_name == 'ecfp':
            return [
                sum([
                    voxelize(
                        convert_atom_to_voxel,
                        xyz,
                        box_width=self.box_width,
                        voxel_width=self.voxel_width,
                        hash_function=hash_ecfp,
                        feature_dict=ecfp_dict,
                        nb_channel=2**self.ecfp_power,
                    ) for xyz, ecfp_dict in zip(
                        (prot_xyz, lig_xyz),
                        featurize_contacts_ecfp(
                            (prot_xyz, prot_rdk), (lig_xyz, lig_rdk),
                            distances,
                            cutoff=self.cutoffs['ecfp_cutoff'],
                            ecfp_degree=self.ecfp_degree))
                ])
            ]
        if feature_name == 'splif':
            return [
                voxelize(
                    convert_atom_pair_to_voxel,
                    (prot_xyz, lig_xyz),
                    box_width=self.box_width,
                    voxel_width=self.voxel_width,
                    hash_function=hash_ecfp_pair,
                    feature_dict=splif_dict,
                    nb_channel=2**self.splif_power,
                ) for splif_dict in featurize_splif((prot_xyz, prot_rdk), (
                    lig_xyz, lig_rdk), self.cutoffs['splif_contact_bins'],
                                                    distances, self.ecfp_degree)
            ]
        if feature_name == 'sybyl':

            def hash_sybyl_func(x):
                hash_sybyl(x, sybyl_types=self.sybyl_types)

            return [
                voxelize(
                    convert_atom_to_voxel,
                    xyz,
                    box_width=self.box_width,
                    voxel_width=self.voxel_width,
                    hash_function=hash_sybyl_func,
                    feature_dict=sybyl_dict,
                    nb_channel=len(self.sybyl_types),
                ) for xyz, sybyl_dict in zip(
                    (prot_xyz, lig_xyz),
                    featurize_binding_pocket_sybyl(
                        prot_xyz,
                        prot_rdk,
                        lig_xyz,
                        lig_rdk,
                        distances,
                        cutoff=self.cutoffs['sybyl_cutoff']))
            ]
        if feature_name == 'salt_bridge':
            return [
                voxelize(
                    convert_atom_pair_to_voxel,
                    (prot_xyz, lig_xyz),
                    box_width=self.box_width,
                    voxel_width=self.voxel_width,
                    feature_list=compute_salt_bridges(
                        prot_rdk,
                        lig_rdk,
                        distances,
                        cutoff=self.cutoffs['salt_bridges_cutoff']),
                    nb_channel=1,
                )
            ]
        if feature_name == 'charge':
            return [
                sum([
                    voxelize(convert_atom_to_voxel,
                             xyz,
                             box_width=self.box_width,
                             voxel_width=self.voxel_width,
                             feature_dict=compute_charge_dictionary(mol),
                             nb_channel=1,
                             dtype="np.float16")
                    for xyz, mol in ((prot_xyz, prot_rdk), (lig_xyz, lig_rdk))
                ])
            ]
        if feature_name == 'hbond':
            return [
                voxelize(
                    convert_atom_pair_to_voxel,
                    (prot_xyz, lig_xyz),
                    box_width=self.box_width,
                    voxel_width=self.voxel_width,
                    feature_list=hbond_list,
                    nb_channel=2**0,
                ) for hbond_list in
                compute_hydrogen_bonds((prot_xyz, prot_rdk), (
                    lig_xyz,
                    lig_rdk), distances, self.cutoffs['hbond_dist_bins'],
                                       self.cutoffs['hbond_angle_cutoffs'])
            ]
        if feature_name == 'pi_stack':
            return voxelize_pi_stack(prot_xyz, prot_rdk, lig_xyz, lig_rdk,
                                     distances,
                                     self.cutoffs['pi_stack_dist_cutoff'],
                                     self.cutoffs['pi_stack_angle_cutoff'],
                                     self.box_width, self.voxel_width)
        if feature_name == 'cation_pi':
            return [
                sum([
                    voxelize(
                        convert_atom_to_voxel,
                        xyz,
                        box_width=self.box_width,
                        voxel_width=self.voxel_width,
                        feature_dict=cation_pi_dict,
                        nb_channel=1,
                    ) for xyz, cation_pi_dict in zip(
                        (prot_xyz, lig_xyz),
                        compute_binding_pocket_cation_pi(
                            prot_rdk,
                            lig_rdk,
                            dist_cutoff=self.cutoffs['cation_pi_dist_cutoff'],
                            angle_cutoff=self.cutoffs['cation_pi_angle_cutoff'],
                        ))
                ])
            ]
        raise ValueError('Unknown feature type "%s"' % feature_name)

    def _featurize(self, complex):
        """Computes grid featurization of protein/ligand complex.

        Takes as input filenames pdb of the protein, pdb of the ligand.

        This function then computes the centroid of the ligand; decrements this
        centroid from the atomic coordinates of protein and ligand atoms, and then
        merges the translated protein and ligand. This combined system/complex is
        then saved.

        This function then computes a featurization with scheme specified by the user.

        Parameters
        ----------
        complex: Tuple[str, str]
            Filenames for molecule and protein.
        """
        try:
            mol_pdb_file, protein_pdb_file = complex
            time1 = time.time()

            protein_xyz, protein_rdk = load_molecule(protein_pdb_file,
                                                     calc_charges=True,
                                                     sanitize=self.sanitize)
            time2 = time.time()
            logger.info(
                "TIMING: Loading protein coordinates took %0.3f s" %
                (time2 - time1), self.verbose)
            time1 = time.time()
            ligand_xyz, ligand_rdk = load_molecule(mol_pdb_file,
                                                   calc_charges=True,
                                                   sanitize=self.sanitize)
            time2 = time.time()
            logger.info(
                "TIMING: Loading ligand coordinates took %0.3f s" %
                (time2 - time1), self.verbose)
        except MoleculeLoadException:
            logger.warning("Some molecules cannot be loaded by Rdkit. Skipping")
            return None

        time1 = time.time()
        centroid = compute_centroid(ligand_xyz)
        ligand_xyz = subtract_centroid(ligand_xyz, centroid)
        protein_xyz = subtract_centroid(protein_xyz, centroid)
        time2 = time.time()
        logger.info(
            "TIMING: Centroid processing took %0.3f s" % (time2 - time1),
            self.verbose)

        pairwise_distances = compute_pairwise_distances(protein_xyz, ligand_xyz)

        transformed_systems = {}
        transformed_systems[(0, 0)] = [protein_xyz, ligand_xyz]

        for i in range(self.nb_rotations):
            rotated_system = rotate_molecules([protein_xyz, ligand_xyz])
            transformed_systems[(i + 1, 0)] = rotated_system

        features_dict = {}
        for system_id, (protein_xyz, ligand_xyz) in transformed_systems.items():
            feature_arrays = []
            for is_flat, function_name in self.feature_types:

                result = self._compute_feature(
                    function_name,
                    protein_xyz,
                    protein_rdk,
                    ligand_xyz,
                    ligand_rdk,
                    pairwise_distances,
                )
                feature_arrays += result

                if self.flatten:
                    features_dict[system_id] = np.concatenate([
                        feature_array.flatten()
                        for feature_array in feature_arrays
                    ])
                else:
                    features_dict[system_id] = np.concatenate(feature_arrays,
                                                              axis=-1)

        features = np.concatenate(list(features_dict.values()))
        return features



"""
hiv dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

HIV_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/HIV.csv"
HIV_TASKS = ["HIV_active"]


class _HIVLoader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "HIV.csv")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=HIV_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_hiv(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['balancing'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load HIV dataset

    The HIV dataset was introduced by the Drug Therapeutics
    Program (DTP) AIDS Antiviral Screen, which tested the ability
    to inhibit HIV replication for over 40,000 compounds.
    Screening results were evaluated and placed into three
    categories: confirmed inactive (CI),confirmed active (CA) and
    confirmed moderately active (CM). We further combine the
    latter two labels, making it a classification task between
    inactive (CI) and active (CA and CM).

    Scaffold splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "smiles": SMILES representation of the molecular structure
    - "activity": Three-class labels for screening results: CI/CM/CA
    - "HIV_active": Binary labels for screening results: 1 (CA/CM) and 0 (CI)

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in

    References
    ----------
    .. [1] AIDS Antiviral Screen Data.
        https://wiki.nci.nih.gov/display/NCIDTPdata/AIDS+Antiviral+Screen+Data
    """
    loader = _HIVLoader(featurizer, splitter, transformers, HIV_TASKS, data_dir,
                        save_dir, **kwargs)
    return loader.load_dataset('hiv', reload)



import torch.nn as nn
import torch
import math
from typing import List, Dict, Optional, Union

from deepchem.models.torch_models import TorchModel
from deepchem.models.losses import L2Loss, SoftmaxCrossEntropy, SigmoidCrossEntropy


class Smiles2Vec(nn.Module):
    """
    Implements the Smiles2Vec model, that learns neural representations of SMILES
    strings which can be used for downstream tasks.

    The goal here is to take SMILES strings as inputs, turn them into vector representations
    which can then be used in predicting molecular properties.

    The model consists of an Embedding layer that retrieves embeddings for each
    character in the SMILES string. These embeddings are learnt jointly with the
    rest of the model. The output from the embedding layer is a tensor of shape
    (batch_size, seq_len, embedding_dim). This tensor can optionally be fed
    through a 1D convolutional layer, before being passed to a series of RNN cells
    (optionally bidirectional). The final output from the RNN cells aims
    to have learnt the temporal dependencies in the SMILES string, and in turn
    information about the structure of the molecule, which is then used for
    molecular property prediction.

    In the paper, the authors also train an explanation mask to endow the model
    with interpretability and gain insights into its decision making. This segment
    is currently not a part of this implementation as this was
    developed for the purpose of investigating a transfer learning protocol,
    ChemNet.

    References
    ----------
    .. [1] Goh et al., "SMILES2vec: An Interpretable General-Purpose Deep Neural Network for
    Predicting Chemical Properties" (https://arxiv.org/pdf/1712.02034.pdf)
    .. [2] Chemnet (https://arxiv.org/abs/1712.02734)
    """

    def __init__(
        self,
        char_to_idx: Dict,
        n_tasks: int = 10,
        max_seq_len: int = 270,
        embedding_dim: int = 50,
        n_classes: int = 2,
        use_bidir: bool = True,
        use_conv: bool = True,
        filters: int = 192,
        kernel_size: int = 3,
        strides: int = 1,
        rnn_sizes: List[int] = [224, 384],
        rnn_types: List[str] = ["GRU", "GRU"],
        mode: str = "regression",
    ):
        """
        Parameters
        ----------
        char_to_idx: dict,
            char_to_idx contains character to index mapping for SMILES characters
        embedding_dim: int, default 50
            Size of character embeddings used.
        use_bidir: bool, default True
            Whether to use BiDirectional RNN Cells
        use_conv: bool, default True
            Whether to use a conv-layer
        kernel_size: int, default 3
            Kernel size for convolutions
        filters: int, default 192
            Number of filters
        strides: int, default 1
            Strides used in convolution
        rnn_sizes: list[int], default [224, 384]
            Number of hidden units in the RNN cells
        mode: str, default regression
            Whether to use model for regression or classification
        """

        super(Smiles2Vec, self).__init__()

        self.char_to_idx = char_to_idx
        self.n_classes = n_classes
        self.max_seq_len = max_seq_len
        self.embedding_dim = embedding_dim
        self.use_bidir = use_bidir
        self.use_conv = use_conv
        if use_conv:
            self.kernel_size = kernel_size
            self.filters = filters
            self.strides = strides
        self.rnn_types = rnn_types
        self.rnn_sizes = rnn_sizes
        assert len(rnn_sizes) == len(
            rnn_types), "Should have same number of hidden units as RNNs"
        self.n_tasks = n_tasks
        self.mode = mode
        self.embedding = nn.Embedding(num_embeddings=len(self.char_to_idx),
                                      embedding_dim=self.embedding_dim)

        if use_conv:
            self.conv1d = nn.Conv1d(embedding_dim,
                                    filters,
                                    kernel_size,
                                    stride=strides)
            self.relu = nn.ReLU()
            self.conv_output = math.floor(
                (max_seq_len - kernel_size + 1) / strides)

        self.RNN_DICT = {'RNN': nn.RNN, 'LSTM': nn.LSTM, 'GRU': nn.GRU}
        self.filters = filters
        self.rnn_sizes = rnn_sizes
        self.use_bidir = use_bidir
        self.rnn_layers: nn.ModuleList = nn.ModuleList()
        self.output_activation: Optional[nn.Module] = None

        # Define RNN layers
        for idx, rnn_type in enumerate(self.rnn_types[:-1]):
            rnn_layer = self.RNN_DICT[rnn_type](
                input_size=self.filters if idx == 0 else self.rnn_sizes[idx -
                                                                        1],
                hidden_size=self.rnn_sizes[idx],
                batch_first=True,
                bidirectional=self.use_bidir)
            self.rnn_layers.append(rnn_layer)

        # Create the last RNN layer separately
        last_layer_input_size = self.rnn_sizes[-2] * (2
                                                      if self.use_bidir else 1)
        last_rnn_layer = self.RNN_DICT[self.rnn_types[-1]]
        self.last_rnn_layer = last_rnn_layer(input_size=last_layer_input_size,
                                             hidden_size=self.rnn_sizes[-1],
                                             batch_first=True,
                                             bidirectional=self.use_bidir)

        input_size = self.rnn_sizes[-1] * (2 if self.use_bidir else 1)

        # Define the fully connected layer for final output
        if self.mode == "classification":
            self.fc = nn.Linear(input_size, self.n_tasks * self.n_classes)
        else:
            self.fc = nn.Linear(input_size, self.n_tasks)

        if self.mode == "classification":
            if self.n_classes == 2:
                self.output_activation = nn.Sigmoid()
            else:
                self.output_activation = nn.Softmax(dim=-1)
        else:
            self.output_activation = None

    def forward(self, smiles_seqs: List):
        """Build the model."""

        rnn_input = self.embedding(smiles_seqs)

        if self.use_conv:
            # Convert to (batch_size, seq_len, filters) for conv1d
            rnn_input = rnn_input.permute(0, 2, 1)
            rnn_input = self.conv1d(rnn_input)
            rnn_input = self.relu(rnn_input)
            rnn_input = rnn_input.permute(
                0, 2, 1)  # Convert back to (batch_size, seq_len, filters)

        # RNN layers
        rnn_embeddings = rnn_input
        for layer in self.rnn_layers:
            rnn_embeddings, _ = layer(rnn_embeddings)

        # Pass through the last RNN layer
        _, hn = self.last_rnn_layer(rnn_embeddings)

        # to match the way tf rnn layers output
        hn = hn.permute(1, 0, 2)
        x = hn.contiguous().view(hn.shape[0], -1)

        Logits = self.fc(x)

        if self.mode == "classification":
            Logits = Logits.view(-1, self.n_tasks, self.n_classes)

            if self.output_activation is not None:
                output = self.output_activation(Logits)
            return Logits, output
        else:
            Logits = Logits.view(-1, self.n_tasks, 1)
            return Logits


class Smiles2VecModel(TorchModel):
    """
    Implements the Smiles2Vec model, that learns neural representations of SMILES
    strings which can be used for downstream tasks.

    The goal here is to take SMILES strings as inputs, turn them into vector representations
    which can then be used in predicting molecular properties.

    The model consists of an Embedding layer that retrieves embeddings for each
    character in the SMILES string. These embeddings are learnt jointly with the
    rest of the model. The output from the embedding layer is a tensor of shape
    (batch_size, seq_len, embedding_dim). This tensor can optionally be fed
    through a 1D convolutional layer, before being passed to a series of RNN cells
    (optionally bidirectional). The final output from the RNN cells aims
    to have learnt the temporal dependencies in the SMILES string, and in turn
    information about the structure of the molecule, which is then used for
    molecular property prediction.

    In the paper, the authors also train an explanation mask to endow the model
    with interpretability and gain insights into its decision making. This segment
    is currently not a part of this implementation as this was
    developed for the purpose of investigating a transfer learning protocol,
    ChemNet.

    Examples
    --------
    >>> import deepchem as dc
    >>> import os
    >>> import numpy as np
    >>> from deepchem.models.torch_models import Smiles2VecModel
    >>> from deepchem.feat import create_char_to_idx, SmilesToSeq

    >>> smiles = ["C1CCC1", "C1=CC=CN=C1"]
    >>> data_points = len(smiles)

    >>> max_seq_len = 20
    >>> max_len = 250
    >>> pad_len = 10
    >>> n_tasks = 5

    >>> dataset_file = os.path.join(os.path.dirname(__file__), "tests", "assets", "chembl_25_small.csv")
    >>> char_to_idx = create_char_to_idx(dataset_file, max_len=max_len, smiles_field="smiles")
    >>> featurizer = SmilesToSeq(char_to_idx=char_to_idx, max_len=max_len, pad_len=pad_len)

    >>> X = featurizer.featurize(smiles)
    >>> y = np.random.normal(size=(data_points, n_tasks))
    >>> dataset = dc.data.NumpyDataset(X=X, y=y)
    >>> w = np.ones(shape=(data_points, n_tasks))

    >>> dataset = dc.data.NumpyDataset(X[:data_points, :max_seq_len], y, w, dataset.ids[:data_points])
    >>> metric = dc.metrics.Metric(dc.metrics.mean_absolute_error, mode="regression")

    >>> model = Smiles2VecModel(char_to_idx=char_to_idx, max_seq_len=max_seq_len, use_conv=True, n_tasks=n_tasks, model_dir=None, mode="regression")
    >>> loss = model.fit(dataset, nb_epoch=10)

    References
    ----------
    .. [1] Goh et al., "SMILES2vec: An Interpretable General-Purpose Deep Neural Network for
    Predicting Chemical Properties" (https://arxiv.org/pdf/1712.02034.pdf)

    .. [2] Using Rule-Based Labels for Weak Supervised Learning: A ChemNet for Transferable
    Chemical Property Prediction(https://arxiv.org/abs/1712.02734)

    """

    def __init__(self,
                 char_to_idx: Dict,
                 n_tasks: int = 10,
                 max_seq_len: int = 270,
                 embedding_dim: int = 50,
                 n_classes: int = 2,
                 use_bidir: bool = True,
                 use_conv: bool = True,
                 filters: int = 192,
                 kernel_size: int = 3,
                 strides: int = 1,
                 rnn_sizes: List[int] = [224, 384],
                 rnn_types: List[str] = ["GRU", "GRU"],
                 mode: str = "regression",
                 device: Optional[torch.device] = None,
                 **kwargs):
        """
        Parameters
        ----------
        char_to_idx: dict,
            char_to_idx contains character to index mapping for SMILES characters
        embedding_dim: int, default 50
            Size of character embeddings used.
        use_bidir: bool, default True
            Whether to use BiDirectional RNN Cells
        use_conv: bool, default True
            Whether to use a conv-layer
        kernel_size: int, default 3
            Kernel size for convolutions
        filters: int, default 192
            Number of filters
        strides: int, default 1
            Strides used in convolution
        rnn_sizes: list[int], default [224, 384]
            Number of hidden units in the RNN cells
        mode: str, default regression
            Whether to use model for regression or classification
        device: torch.device, optional (default None)
            the device on which to run computations.  If None, a device is
            chosen automatically
        """
        self.n_tasks = n_tasks
        self.n_classes = n_classes
        self.mode: str = mode
        self.model = Smiles2Vec(char_to_idx=char_to_idx,
                                n_tasks=n_tasks,
                                max_seq_len=max_seq_len,
                                embedding_dim=embedding_dim,
                                n_classes=n_classes,
                                use_bidir=use_bidir,
                                use_conv=use_conv,
                                filters=filters,
                                kernel_size=kernel_size,
                                strides=strides,
                                rnn_sizes=rnn_sizes,
                                rnn_types=rnn_types,
                                mode=mode)
        loss: Union[SigmoidCrossEntropy, SoftmaxCrossEntropy, L2Loss]

        if mode == "classification":
            if n_classes == 2:
                loss = SigmoidCrossEntropy()
            else:
                loss = SoftmaxCrossEntropy()

            output_types = ['prediction', 'loss']

        else:
            loss = L2Loss()
            output_types = ['prediction']

        super(Smiles2VecModel, self).__init__(model=self.model,
                                              loss=loss,
                                              output_types=output_types,
                                              device=device,
                                              **kwargs)



"""
Test for DFT Config Utilities
"""

from deepchem.utils.dft_utils.config import config


def test_config():
    conf = config
    assert conf.THRESHOLD_MEMORY == 10 * 1024**3
    assert conf.CHUNK_MEMORY == 16 * 1024**2
    assert conf.VERBOSE == 0



import contextlib
import warnings
from typing import Any, Callable, Dict, List, Optional, Tuple
import torch
import numpy as np
try:
    import h5py
except:
    warnings.warn("h5py is not installed, cache will not work.")


class Cache(object):
    """
    Internal object to store/load cache for heavy-load calculations.
    For a class to be able to use cache, it has to have cache as its members.
    The same cache object can also be passed to other object either via
        `.add_prefix` method or just passed as it is.
    The list of cacheable parameters are constructed within the class.
    The file to store cache also set within the class via `set` method, if no
        file is set, then the cache does not work.
    To load/store cache, the context handler, `.open()` needs to be called.
    We use the magic method `cache` here to load/store cache, i.e. if the given
        dataset is not in the cache, then it is calculated then stored.
    Otherwise, the dataset is just loaded from the cache file.
    To avoid loading the wrong cache (e.g. the same file name is set for two
        different objects), we provide `check_signature` method which will
        raise a warning if the signature from the cached file and the input of
        `check_signature` do not match.

    Examples
    --------
    >>> class A:
    ...     def __init__(self):
    ...         self.cache = Cache.get_dummy()
    ...     def foo(self, x):
    ...         return self.cache.cache("foo", lambda: x * x)
    >>> a = A()
    >>> a.foo(2)
    4
    >>> a.foo(2)
    4

    """

    def __init__(self):
        """Initialize the cache object."""
        self._cacheable_pnames: List[str] = []
        self._fname: Optional[str] = None
        self._pnames_to_cache: Optional[List[str]] = None
        self._fhandler: Optional[h5py.File] = None

    def set(self, fname: str, pnames: Optional[List[str]] = None):
        """set up the cache

        Parameters
        ----------
        fname : str
            File name to store the cache
        pnames : Optional[List[str]], optional
            List of parameter names to be cached, by default None

        Raises
        ------
        RuntimeError
            If the cache has been set before
        """
        self._fname = fname
        self._pnames_to_cache = pnames

    def cache(self, pname: str, fcn: Callable[[],
                                              torch.Tensor]) -> torch.Tensor:
        """cache the result of the function

        Parameters
        ----------
        pname : str
            Parameter name to be cached
        fcn : Callable[[], torch.Tensor]
            Function to be cached

        Returns
        -------
        torch.Tensor
            Result of the function

        """
        # if not has been set, then just calculate and return
        if not self.isset():
            return fcn()

        # if the param is not to be cached, then just calculate and return
        if not self._pname_to_cache(pname):
            return fcn()

        # get the dataset name
        dset_name = self._pname2dsetname(pname)

        # check if dataset name is in the file (cache exist)
        file = self._get_file_handler()
        if dset_name in file:
            # retrieve the dataset if it has been computed before
            res = self._load_dset(dset_name, file)
        else:
            # if not in the file, then compute the tensor and save the cache
            res = fcn()
            self._save_dset(dset_name, file, res)
        return res

    def cache_multi(self, pnames: List[str], fcn: Callable[[], Tuple[torch.Tensor, ...]]) \
            -> Tuple[torch.Tensor, ...]:
        """Cache the result of the function for multiple parameters

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached
        fcn : Callable[[], Tuple[torch.Tensor, ...]]
            Function to be cached

        Returns
        -------
        Tuple[torch.Tensor, ...]
            Tuple of results of the function

        """
        if not self.isset():
            return fcn()

        # if any of the pnames not to be cached, then just calculate and return
        if not all([self._pname_to_cache(pname) for pname in pnames]):
            return fcn()

        # get the dataset names
        dset_names = [self._pname2dsetname(pname) for pname in pnames]

        # get the file handler
        file = self._get_file_handler()

        # check if all the dataset is in the cache file, otherwise, just evaluate
        all_dset_names_in_file = all(
            [dset_name in file for dset_name in dset_names])
        if all_dset_names_in_file:
            all_res = tuple(
                self._load_dset(dset_name, file) for dset_name in dset_names)
        else:
            all_res = fcn()
            for dset_name, res in zip(dset_names, all_res):
                self._save_dset(dset_name, file, res)
        return all_res

    @contextlib.contextmanager
    def open(self):
        """Open the cache file

        Yields
        -------
        h5py.File
            File handler

        """
        try:
            if self._fname is not None:
                self._fhandler = h5py.File(self._fname, "a")
            yield self._fhandler
        finally:
            if self._fname is not None:
                self._fhandler.close()
                self._fhandler = None

    def add_prefix(self, prefix: str) -> 'Cache':
        """Return a Cache object that will add the prefix for every input of
        parameter names

        Parameters
        ----------
        prefix : str
            Prefix to be added to the parameter names

        Returns
        -------
        Cache
            Cache object with the prefix added to the parameter names

        """
        prefix = normalize_prefix(prefix)
        return _PrefixedCache(self, prefix)

    def add_cacheable_params(self, pnames: List[str]):
        """Add the cacheable parameter names

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached

        """
        self._cacheable_pnames.extend(pnames)

    def get_cacheable_params(self) -> List[str]:
        """return the cacheable parameter names

        Returns
        -------
        List[str]
            List of cacheable parameter names

        """
        return self._cacheable_pnames

    def check_signature(self,
                        sig_dict: Dict[str, Any],
                        _groupname: Optional[str] = "/"):
        """Executed while the file is opened, if there is no signature, then add
        the signature, if there is a signature in the file, then match it.
        If they do not match, print a warning, otherwise, do nothing
        _groupname is internal parameter within this class, should not be
         specified by classes other than Cache and its inheritances.

        Parameters
        ----------
        sig_dict : Dict[str, Any]
            Dictionary of signature
        _groupname : Optional[str] (default "/")
            Group name in the h5py file

        """
        if not self.isset():
            return

        fh = self._get_file_handler()
        sig_attrname = "signature"

        # get the group
        if _groupname in fh:
            group = fh[_groupname]
        else:
            group = fh.create_group(_groupname)

        # create the signature string
        sig_str = "\n\n".join(
            ["%s:\n%s" % (k, str(v)) for (k, v) in sig_dict.items()])

        # if the signature does not exist, then write the signature as an attribute
        if sig_attrname not in group.attrs:
            group.attrs[sig_attrname] = str(sig_str)

        # if the signature exist, then check it, if it does not match, raise a warning
        else:
            cached_sig = str(group.attrs[sig_attrname])
            if cached_sig != sig_str:
                msg = (
                    "Mismatch of the cached signature.\nCached signature:\n%s\n"
                    "-----------------------\n"
                    "Current signature:\n%s" % (cached_sig, sig_str))
                warnings.warn(msg)

        group.attrs["signature"] = str(sig_str)

    @staticmethod
    def get_dummy() -> 'Cache':
        """Returns a dummy cache that does not do anything

        Returns
        -------
        Cache
            Dummy cache object

        """
        return _DummyCache()

    def _pname_to_cache(self, pname: str) -> bool:
        """Check if the input parameter name is to be cached

        Parameters
        ----------
        pname : str
            Parameter name to be checked

        Returns
        -------
        bool
            Indicator whether the parameter name is to be cached

        """
        return (self._pnames_to_cache is None) or (pname
                                                   in self._pnames_to_cache)

    def _pname2dsetname(self, pname: str) -> str:
        """Convert the parameter name to dataset name

        Parameters
        ----------
        pname : str
            Parameter name to be converted to dataset name

        Returns
        -------
        str
            Dataset name

        """
        return pname.replace(".", "/")

    def _get_file_handler(self) -> h5py.File:
        """Return the file handler, if the file is not opened yet,
        then raise an error

        Returns
        -------
        h5py.File
            File handler

        """
        if self._fhandler is None:
            msg = "The cache file has not been opened yet, please use .open() before reading/writing to the cache"
            raise RuntimeError(msg)
        else:
            return self._fhandler

    def isset(self) -> bool:
        """Returns the indicator whether the cache object has been set

        Returns
        -------
        bool
            Indicator whether the cache object has been set

        """
        return self._fname is not None

    def _load_dset(self, dset_name: str, fhandler: h5py.File) -> torch.Tensor:
        """Load the dataset from the file handler (check is performed outside)

        Parameters
        ----------
        dset_name : str
            Dataset name
        fhandler : h5py.File
            File handler for the dataset

        """
        dset_np = np.asarray(fhandler[dset_name])
        dset = torch.as_tensor(dset_np)
        return dset

    def _save_dset(self, dset_name: str, fhandler: h5py.File,
                   dset: torch.Tensor):
        """Save res to the h5py in the dataset name

        Parameters
        ----------
        dset_name : str
            Dataset name
        fhandler : h5py.File
            File handler for the dataset
        dset : torch.Tensor
            Tensor to be saved

        """
        fhandler[dset_name] = dset.detach()


class _PrefixedCache(Cache):
    """This class adds a prefix to every parameter names input

    Examples
    --------
    >>> cache = Cache.get_dummy()
    >>> pcache = _PrefixedCache(cache, "prefix.")
    >>> pcache.cache("foo", lambda: 1)
    1
    >>> pcache.cache("foo", lambda: 2)
    2

    """

    def __init__(self, obj: Cache, prefix: str):
        """Initialize the PrefixedCache object

        Parameters
        ----------
        obj : Cache
            Cache object to be prefixed
        prefix : str
            Prefix to be added to the parameter names

        """
        self._obj = obj
        self._prefix = prefix

    def set(self, fname: str, pnames: Optional[List[str]] = None):
        """set must only be done in the parent object, not in the children objects

        Parameters
        ----------
        fname : str
            File name to store the cache
        pnames : Optional[List[str]] (default None)
            List of parameter names to be cached, by default None

        """
        raise RuntimeError("Cache.set() must be done on non-prefixed cache")

    def cache(self, pname: str, fcn: Callable[[],
                                              torch.Tensor]) -> torch.Tensor:
        """Cache the result of the function

        Parameters
        ----------
        pname : str
            Parameter name to be cached
        fcn : Callable[[], torch.Tensor]
            Function to be cached

        Returns
        -------
        torch.Tensor
            Result of the function

        """
        return self._obj.cache(self._prefixed(pname), fcn)

    def cache_multi(self, pnames: List[str], fcn: Callable[[], Tuple[torch.Tensor, ...]]) \
            -> Tuple[torch.Tensor, ...]:
        """Cache the result of the function for multiple parameters

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached
        fcn : Callable[[], Tuple[torch.Tensor, ...]]
            Function to be cached

        Returns"""
        ppnames = [self._prefixed(pname) for pname in pnames]
        return self._obj.cache_multi(ppnames, fcn)

    @contextlib.contextmanager
    def open(self):
        """Open the cache file

        Yields
        -------
        h5py.File
            File handler

        """
        with self._obj.open() as f:
            try:
                yield f
            finally:
                pass

    def add_prefix(self, prefix: str) -> Cache:
        """Return a deeper prefixed object

        Parameters
        ----------
        prefix : str
            Prefix to be added to the parameter names

        Returns
        -------
        Cache
            Cache object with the prefix added to the parameter names

        """
        prefix = self._prefixed(normalize_prefix(prefix))
        return self._obj.add_prefix(prefix)

    def add_cacheable_params(self, pnames: List[str]):
        """Add the cacheable parameter names

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached

        """
        pnames = [self._prefixed(pname) for pname in pnames]
        self._obj.add_cacheable_params(pnames)

    def get_cacheable_params(self) -> List[str]:
        """This can only be done on the root cache (non-prefixed) to avoid
        confusion about which name should be provided (absolute or relative)

        Returns
        -------
        List[str]
            List of cacheable parameter names

        Raises
        ------
        RuntimeError
            If the method is called on the prefixed cache

        """
        raise RuntimeError(
            "Cache.get_cacheable_params() must be done on non-prefixed cache")

    def check_signature(self,
                        sig_dict: Dict[str, Any],
                        _groupname: Optional[str] = "/"):
        """use the prefix as the groupname to do signature check of the root object

        Parameters
        ----------
        sig_dict : Dict[str, Any]
            Dictionary of signature
        _groupname : Optional[str] (default "/")
            Group name in the h5py file

        """
        if not self.isset():
            return

        groupname = "/" + self._prefix.replace(".", "/")
        if groupname.endswith("/"):
            groupname = groupname[:-1]

        self._obj.check_signature(sig_dict, _groupname=groupname)

    def isset(self) -> bool:
        """Returns the indicator whether the cache object has been set

        Returns
        -------
        bool
            Indicator whether the cache object has been set

        """
        return self._obj.isset()

    def _prefixed(self, pname: str) -> str:
        """Returns the prefixed name

        Parameters
        ----------
        pname : str
            Parameter name to be prefixed

        Returns
        -------
        str
            Prefixed parameter name

        """
        return self._prefix + pname


class _DummyCache(Cache):
    """This class just an interface of cache without doing anything.

    Examples
    --------
    >>> cache = _DummyCache()
    >>> cache.cache("foo", lambda: 1)
    1
    >>> cache.cache("foo", lambda: 2)
    2

    """

    def __init__(self):
        """Initialize the DummyCache object."""
        pass

    def set(self, fname: str, pnames: Optional[List[str]] = None):
        """set up the cache

        Parameters
        ----------
        fname : str
            File name to store the cache
        pnames : Optional[List[str]], optional
            List of parameter names to be cached, by default None

        """
        pass

    def cache(self, pname: str, fcn: Callable[[],
                                              torch.Tensor]) -> torch.Tensor:
        """cache the result of the function

        Parameters
        ----------
        pname : str
            Parameter name to be cached
        fcn : Callable[[], torch.Tensor]
            Function to be cached

        Returns
        -------
        torch.Tensor
            Result of the function

        """
        return fcn()

    def cache_multi(self, pnames: List[str], fcn: Callable[[], Tuple[torch.Tensor, ...]]) \
            -> Tuple[torch.Tensor, ...]:
        """Cache the result of the function for multiple parameters

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached
        fcn : Callable[[], Tuple[torch.Tensor, ...]]
            Function to be cached

        Returns
        -------
        Tuple[torch.Tensor, ...]
            Tuple of results of the function

        """
        return fcn()

    @contextlib.contextmanager
    def open(self):
        """Open the cache file

        Yields
        -------
        h5py.File
            File handler

        """
        try:
            yield None
        finally:
            pass

    def add_prefix(self, prefix: str) -> Cache:
        """Return a Cache object that will add the prefix for every input of
        parameter names

        Parameters
        ----------
        prefix : str
            Prefix to be added to the parameter names

        Returns
        -------
        Cache
            Cache object with the prefix added to the parameter names

        """
        return self

    def add_cacheable_params(self, pnames: List[str]):
        """Add the cacheable parameter names

        Parameters
        ----------
        pnames : List[str]
            List of parameter names to be cached

        """
        pass

    def get_cacheable_params(self) -> List[str]:
        """return the cacheable parameter names

        Returns
        -------
        List[str]
            List of cacheable parameter names

        """
        return []

    def check_signature(self,
                        sig_dict: Dict[str, Any],
                        _groupname: Optional[str] = "/"):
        """Executed while the file is opened, if there is no signature, then add
        the signature, if there is a signature in the file, then match it.
        If they do not match, print a warning, otherwise, do nothing
        _groupname is internal parameter within this class, should not be
         specified by classes other than Cache and its inheritances.

        Parameters
        ----------
        sig_dict : Dict[str, Any]
            Dictionary of signature
        _groupname : Optional[str] (default "/")
            Group name in the h5py file

        """
        pass

    def isset(self):
        """Returns the indicator whether the cache object has been set

        Returns
        -------
        bool
            Indicator whether the cache object has been set

        """
        return False


def normalize_prefix(prefix: str) -> str:
    """Added a dot at the end of prefix if it is not so.

    Examples
    --------
    >>> _normalize_prefix("prefix")
    'prefix.'
    >>> _normalize_prefix("prefix.")
    'prefix.'

    Parameters
    ----------
    prefix : str
        Prefix to be normalized

    Returns
    -------
    str
        Normalized prefix
    """
    if not prefix.endswith("."):
        prefix = prefix + "."
    return prefix



import torch
import torch.nn as nn
import torch.nn.functional as F

from deepchem.data import Dataset
from deepchem.models import losses
from deepchem.models.torch_models.torch_model import TorchModel
from deepchem.utils.pytorch_utils import get_activation
from deepchem.utils.typing import OneOrMany, ActivationFn, LossFn

import logging
from collections.abc import Sequence as SequenceCollection
from typing import List, Tuple, Callable, Literal, Optional, Union

logger = logging.getLogger(__name__)


class ProgressiveMultitask(nn.Module):
    """Implements a progressive multitask neural network in PyTorch.

    Progressive networks allow for multitask learning where each task
    gets a new column of weights and lateral connections to previous tasks
    are added to the network. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    Examples
    --------
    >>> import torch
    >>> import deepchem as dc
    >>> n_tasks = 4
    >>> n_features = 1024
    >>> n_classes = 2
    >>> sample = torch.randn(16, n_features)
    >>> model = dc.models.torch_models.ProgressiveMultitask(n_tasks=n_tasks, n_features=n_features, layer_sizes=[1024, 1024], mode="classification", n_classes=n_classes)
    >>> output = model(sample)
    >>> print(output[0].type(), output[1].type())
    torch.FloatTensor torch.FloatTensor
    >>> print(output[0].shape, output[1].shape)
    torch.Size([16, 4, 2]) torch.Size([16, 4, 2])

    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: List[int] = [1000],
                 mode: Literal['regression', 'classification'] = 'regression',
                 alpha_init_stddevs: OneOrMany[float] = 0.02,
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 dropouts: OneOrMany[float] = 0.5,
                 n_classes: int = 1):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        n_features: int
            Size of input feature vector.
        layer_sizes: list of ints
            List of layer sizes.
        mode: str
            Type of model.  Must be 'regression' or 'classification'.
        alpha_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            alpha parameters.
        weight_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            weight parameters.
        bias_init_consts: float or list of floats
            Constant value to initialize bias parameters.
        weight_decay_penalty: float
            Amount of weight decay penalty to use.
        weight_decay_penalty_type: str
            Type of weight decay penalty.  Must be 'l1' or 'l2'.
        activation_fns: str or list of str
            Name of activation function(s) to use.
        dropouts: float or list of floats
            Dropout probability.
        n_classes: int
            The number of classes to predict per task.
        """
        if weight_decay_penalty != 0.0:
            raise ValueError("Weight decay is not currently supported")

        n_layers = len(layer_sizes)
        if not isinstance(weight_init_stddevs, SequenceCollection):
            weight_init_stddevs = [weight_init_stddevs] * n_layers
        if not isinstance(alpha_init_stddevs, SequenceCollection):
            alpha_init_stddevs = [alpha_init_stddevs] * n_layers
        if not isinstance(bias_init_consts, SequenceCollection):
            bias_init_consts = [bias_init_consts] * n_layers
        if not isinstance(dropouts, SequenceCollection):
            dropouts = [dropouts] * n_layers
        if isinstance(
                activation_fns,
                str) or not isinstance(activation_fns, SequenceCollection):
            activation_fns = [activation_fns] * n_layers

        self.mode = mode
        self.n_tasks: int = n_tasks
        self.n_features: int = n_features
        self.layer_sizes: List[int] = layer_sizes
        self.n_classes: int = n_classes
        self.weight_init_stddevs: SequenceCollection[
            float] = weight_init_stddevs
        self.alpha_init_stddevs: SequenceCollection[float] = alpha_init_stddevs
        self.bias_init_consts: SequenceCollection[float] = bias_init_consts
        self.dropouts: SequenceCollection[float] = dropouts
        self.activation_fns: SequenceCollection[Callable] = [
            get_activation(f) for f in activation_fns
        ]

        super(ProgressiveMultitask, self).__init__()

        self.layers: nn.ModuleList = nn.ModuleList()
        self.adapters: nn.ModuleList = nn.ModuleList()
        self.alphas: nn.ModuleList = nn.ModuleList()

        for task in range(n_tasks):
            layer_list = []
            adapter_list = []
            alpha_list = []
            prev_size = n_features
            for i, size in enumerate(self.layer_sizes):
                layer_list.append(self._init_linear(prev_size, size, i))

                if task > 0:
                    if i > 0:
                        adapter, alpha = self._get_adapter(
                            task, prev_size, size, i)
                        adapter_list.append(adapter)
                        alpha_list.append(alpha)

                prev_size = size

            layer_list.append(
                self._init_linear(prev_size, self.n_classes,
                                  len(self.layer_sizes)))
            self.layers.append(nn.ModuleList(layer_list))
            if task > 0:
                adapter, alpha = self._get_adapter(task, prev_size,
                                                   self.n_classes,
                                                   len(self.layer_sizes))
                adapter_list.append(adapter)
                alpha_list.append(alpha)

                self.adapters.append(nn.ModuleList(adapter_list))
                self.alphas.append(nn.ParameterList(alpha_list))

    def _get_adapter(self, task: int, prev_size: int, size: int,
                     layer_num: int) -> Tuple[nn.ModuleList, torch.Tensor]:
        """Creates the adapter layer between previous tasks and the current layer.

        Parameters
        ----------
        task: int
            Task number.
        prev_size: int
            Size of previous layer.
        size: int
            Size of current layer.
        layer_num: int
            Layer number.

        Returns
        -------
        adapter: nn.Sequential
            Adapter layer.
        alpha: torch.Tensor
            Alpha parameter.
        """
        adapter = nn.ModuleList([
            self._init_linear(prev_size * task, prev_size, layer_num),
            self._init_linear(prev_size, size, layer_num, use_bias=False),
        ])
        alpha_init_stddev = (self.alpha_init_stddevs[layer_num]
                             if layer_num < len(self.layer_sizes) else
                             self.alpha_init_stddevs[-1])
        alpha = torch.empty(1, requires_grad=True)
        nn.init.trunc_normal_(alpha, std=alpha_init_stddev)
        return adapter, alpha

    def _init_linear(self,
                     in_features: int,
                     out_features: int,
                     layer_num: int,
                     use_bias: bool = True) -> nn.Linear:
        """Initialises nn.Linear layer weight and bias parameters.

        Parameters
        ----------
        in_features: int
            Size of input feature vector.
        out_features: int
            Size of output feature vector.
        layer_num: int
            Layer number.
        use_bias: bool
            Whether to use bias for Linear layer. Default to True.

        Returns
        -------
        layer: nn.Linear
            Linear layer with initialised parameters.
        """
        if layer_num < len(self.layer_sizes):
            weight_init_stddev = self.weight_init_stddevs[layer_num]
            bias_init_const = self.bias_init_consts[layer_num]
        elif layer_num == len(self.layer_sizes):
            weight_init_stddev = self.weight_init_stddevs[-1]
            bias_init_const = self.bias_init_consts[-1]

        layer = nn.Linear(in_features, out_features, bias=use_bias)
        nn.init.trunc_normal_(layer.weight, std=weight_init_stddev)

        if use_bias:
            nn.init.constant_(layer.bias, bias_init_const)

        return layer

    def forward(
        self, x: torch.Tensor
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Forward pass through the network.

        Parameters
        ----------
        x: torch.Tensor
            Input tensor.

        Returns
        -------
        torch.Tensor
            The Model output tensor of shape (batch_size, n_tasks, n_outputs).

        * When self.mode = `regression`,
            It consists of the output of each task.
        * When self.mode = `classification`,
            It consists of the probability of each class for each task.

        torch.Tensor, optional
            This is only returned when self.mode = `classification`, the output consists of the
            logits for classes before softmax.
        """
        task_outputs: List[torch.Tensor] = []
        layer_logits: List[List[torch.Tensor]] = []
        for task in range(self.n_tasks):
            x_ = x
            layer_outputs = []
            for i, (layer, activation_fn, dropout) in enumerate(
                    zip(self.layers[task], self.activation_fns, self.dropouts)):
                x_ = layer(x_)

                if task > 0 and i > 0:
                    adapter_out = self._get_lateral_contrib(
                        layer_logits, task, i)
                    x_ = x_ + adapter_out

                if i < len(self.layer_sizes):
                    x_ = activation_fn(x_)
                    if dropout > 0.0:
                        x_ = F.dropout(x_, p=dropout, training=self.training)

                layer_outputs.append(x_)

            out_layer_idx = len(self.layer_sizes)
            x_ = self.layers[task][out_layer_idx](x_)  # type: ignore
            if task > 0:
                adapter_out = self._get_lateral_contrib(layer_logits, task,
                                                        out_layer_idx)
                x_ = x_ + adapter_out

            layer_logits.append(layer_outputs)
            task_outputs.append(x_)

        output = torch.stack(task_outputs, dim=1)

        if self.mode == 'classification':
            if self.n_tasks == 1:
                logits = output.view(-1, self.n_classes)
                softmax_dim = 1
            else:
                logits = output.view(-1, self.n_tasks, self.n_classes)
                softmax_dim = 2
            proba = F.softmax(logits, dim=softmax_dim)
            return proba, logits
        else:
            return output

    def _get_lateral_contrib(self, logits, task, layer):
        adapter = self.adapters[task - 1][layer - 1]
        alpha = self.alphas[task - 1][layer - 1]
        prev_logits = [logits[t][layer - 1] for t in range(task)]
        adapter_input = torch.cat(prev_logits, dim=-1)
        adapter_input = alpha * adapter_input

        adapter_out = adapter[0](adapter_input)
        adapter_out = self.activation_fns[layer - 1](adapter_out)
        adapter_out = adapter[1](adapter_out)
        return adapter_out


class ProgressiveMultitaskModel(TorchModel):
    """Implements a progressive multitask neural network in PyTorch.

    Progressive networks allow for multitask learning where each task
    gets a new column of weights and lateral connections to previous tasks
    are added to the network. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    Examples
    --------
    >>> import deepchem as dc
    >>> from deepchem.models.torch_models import ProgressiveMultitaskModel
    >>> featurizer = dc.feat.CircularFingerprint(size=1024, radius=4)
    >>> tasks, datasets, transformers = dc.molnet.load_tox21(featurizer=featurizer)
    >>> train_dataset, valid_dataset, test_dataset = datasets
    >>> n_tasks = len(tasks)
    >>> model = ProgressiveMultitaskModel(n_tasks, 1024, layer_sizes=[1024], mode='classification')
    >>> model.fit(train_dataset, nb_epoch=10)

    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks: int,
                 n_features: int,
                 layer_sizes: List[int] = [1000],
                 mode: Literal['regression', 'classification'] = 'regression',
                 alpha_init_stddevs: OneOrMany[float] = 0.02,
                 weight_init_stddevs: OneOrMany[float] = 0.02,
                 bias_init_consts: OneOrMany[float] = 1.0,
                 weight_decay_penalty: float = 0.0,
                 weight_decay_penalty_type: str = "l2",
                 activation_fns: OneOrMany[ActivationFn] = 'relu',
                 dropouts: OneOrMany[float] = 0.5,
                 n_classes: Optional[int] = None,
                 n_outputs: Optional[int] = None,
                 **kwargs):
        """
        Parameters
        ----------
        n_tasks: int
            Number of tasks.
        n_features: int
            Size of input feature vector.
        layer_sizes: list of ints
            List of layer sizes.
        mode: str
            Type of model.  Must be 'regression' or 'classification'.
        alpha_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            alpha parameters.
        weight_init_stddevs: float or list of floats
            Standard deviation for truncated normal distribution to initialize
            weight parameters.
        bias_init_consts: float or list of floats
            Constant value to initialize bias parameters.
        weight_decay_penalty: float
            Amount of weight decay penalty to use.
        weight_decay_penalty_type: str
            Type of weight decay penalty.  Must be 'l1' or 'l2'.
        activation_fns: str or list of str
            Name of activation function(s) to use.
        dropouts: float or list of floats
            Dropout probability.
        n_classes: int
            The number of classes to predict per task. Default to 2 for classification and 1 for regression.
        n_outputs: int
            The number of outputs to predict per task. Deprecated, use n_classes instead.
        """

        if n_outputs is not None:
            logger.warning(
                "n_outputs is deprecated and will be removed in future versions. Use n_classes instead."
            )
            if n_classes is not None and n_classes != n_outputs:
                raise ValueError(
                    "n_classes and n_outputs should have the same value if both are specified."
                )
            n_classes = n_outputs

        loss: losses.Loss
        if mode == 'regression':
            loss = losses.L2Loss()
            output_types = ['prediction']
            if n_classes is None:
                n_classes = 1
        elif mode == 'classification':
            loss = losses.SparseSoftmaxCrossEntropy()
            output_types = ['prediction', 'loss']
            if n_classes is None:
                n_classes = 2
        else:
            raise ValueError(f'Invalid mode: {mode}')

        model = ProgressiveMultitask(
            n_tasks=n_tasks,
            n_features=n_features,
            layer_sizes=layer_sizes,
            mode=mode,
            alpha_init_stddevs=alpha_init_stddevs,
            weight_init_stddevs=weight_init_stddevs,
            bias_init_consts=bias_init_consts,
            weight_decay_penalty=weight_decay_penalty,
            weight_decay_penalty_type=weight_decay_penalty_type,
            activation_fns=activation_fns,
            dropouts=dropouts,
            n_classes=n_classes)

        super(ProgressiveMultitaskModel,
              self).__init__(model, loss, output_types=output_types, **kwargs)

    @property
    def layers(self):
        return self.model.layers

    @property
    def adapters(self):
        return self.model.adapters

    @property
    def alphas(self):
        return self.model.alphas

    def fit(self,
            dataset: Dataset,
            nb_epoch: int = 10,
            max_checkpoints_to_keep: int = 5,
            checkpoint_interval: int = 1000,
            deterministic: bool = False,
            restore: bool = False,
            variables: Optional[List[torch.nn.Parameter]] = None,
            loss: Optional[LossFn] = None,
            callbacks: Union[Callable, List[Callable]] = [],
            all_losses: Optional[List[float]] = None):

        for task in range(self.model.n_tasks):
            self.fit_task(dataset=dataset,
                          task=task,
                          nb_epoch=nb_epoch,
                          max_checkpoints_to_keep=max_checkpoints_to_keep,
                          checkpoint_interval=checkpoint_interval,
                          deterministic=deterministic,
                          restore=restore,
                          variables=variables,
                          loss=loss,
                          callbacks=callbacks,
                          all_losses=all_losses)

    def fit_task(self,
                 dataset: Dataset,
                 task: int,
                 nb_epoch: int = 10,
                 max_checkpoints_to_keep: int = 5,
                 checkpoint_interval: int = 1000,
                 deterministic: bool = False,
                 restore: bool = False,
                 variables: Optional[List[torch.nn.Parameter]] = None,
                 loss: Optional[LossFn] = None,
                 callbacks: Union[Callable, List[Callable]] = [],
                 all_losses: Optional[List[float]] = None):
        """Train this model on one task. Called by fit() to train each task sequentially.
        Calls fit_generator() internally.

        Parameters
        ----------
        dataset: Dataset
            the Dataset to train on
        task: int
            the task to train on
        nb_epoch: int
            the number of epochs to train for
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        deterministic: bool
            if True, the samples are processed in order.  If False, a different random
            order is used for each epoch.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of torch.nn.Parameter
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step) that will be invoked after
            every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.

        Returns
        -------
        The average loss over the most recent checkpoint interval
        """

        generator = self.default_generator(dataset,
                                           epochs=nb_epoch,
                                           deterministic=deterministic)

        variables = list(self.model.layers[task].parameters())

        if task > 0:
            variables += list(self.model.adapters[task - 1].parameters())
            variables += list(self.model.alphas[task - 1].parameters())

        self.fit_generator(generator=generator,
                           max_checkpoints_to_keep=max_checkpoints_to_keep,
                           checkpoint_interval=checkpoint_interval,
                           restore=restore,
                           variables=variables,
                           loss=loss,
                           callbacks=callbacks,
                           all_losses=all_losses)


class ProgressiveMultitaskRegressor(ProgressiveMultitaskModel):
    """Implements a progressive multitask neural network for regression.

    Progressive Networks: https://arxiv.org/pdf/1606.04671v3.pdf

    Progressive networks allow for multitask learning where each task
    gets a new column of weights. As a result, there is no exponential
    forgetting where previous tasks are ignored.

    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 layer_sizes=[1000],
                 alpha_init_stddevs=0.02,
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 activation_fns='relu',
                 dropouts=0.5,
                 n_outputs=1,
                 n_classes=1,
                 **kwargs):
        """Creates a progressive network.

        Only listing parameters specific to progressive networks here.

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_features: int
            Number of input features
        alpha_init_stddevs: list
            List of standard-deviations for alpha in adapter layers.
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes)+1.  The final element corresponds to the output layer.
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes)+1.
            The final element corresponds to the output layer.  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        """

        if not isinstance(activation_fns, str):
            logger.warning(
                "Activation function should be a string matching an attribute in torch.nn.functional corresponding to an activation function. Using 'relu' as default."
            )
            activation_fns = 'relu'

        super().__init__(n_tasks=n_tasks,
                         n_features=n_features,
                         layer_sizes=layer_sizes,
                         mode='regression',
                         alpha_init_stddevs=alpha_init_stddevs,
                         weight_init_stddevs=weight_init_stddevs,
                         bias_init_consts=bias_init_consts,
                         weight_decay_penalty=weight_decay_penalty,
                         weight_decay_penalty_type=weight_decay_penalty_type,
                         activation_fns=activation_fns,
                         dropouts=dropouts,
                         n_outputs=n_outputs,
                         n_classes=n_classes,
                         **kwargs)


class ProgressiveMultitaskClassifier(ProgressiveMultitaskModel):
    """Implements a progressive multitask neural network for classification.

    Progressive Networks: https://arxiv.org/pdf/1606.04671v3.pdf

    Progressive networks allow for multitask learning where each task
    gets a new column of weights. As a result, there is no exponential
    forgetting where previous tasks are ignored.


    References
    ----------
    See [1]_ for a full description of the progressive architecture

    .. [1] Rusu, Andrei A., et al. "Progressive neural networks." arXiv preprint
        arXiv:1606.04671 (2016).
    """

    def __init__(self,
                 n_tasks,
                 n_features,
                 layer_sizes=[1000],
                 alpha_init_stddevs=0.02,
                 weight_init_stddevs=0.02,
                 bias_init_consts=1.0,
                 weight_decay_penalty=0.0,
                 weight_decay_penalty_type="l2",
                 activation_fns='relu',
                 dropouts=0.5,
                 n_outputs=2,
                 n_classes=2,
                 **kwargs):
        """Creates a progressive network.

        Only listing parameters specific to progressive networks here.

        Parameters
        ----------
        n_tasks: int
            Number of tasks
        n_features: int
            Number of input features
        alpha_init_stddevs: list
            List of standard-deviations for alpha in adapter layers.
        layer_sizes: list
            the size of each dense layer in the network.  The length of this list determines the number of layers.
        weight_init_stddevs: list or float
            the standard deviation of the distribution to use for weight initialization of each layer.  The length
            of this list should equal len(layer_sizes)+1.  The final element corresponds to the output layer.
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        bias_init_consts: list or float
            the value to initialize the biases in each layer to.  The length of this list should equal len(layer_sizes)+1.
            The final element corresponds to the output layer.  Alternatively this may be a single value instead of a list,
            in which case the same value is used for every layer.
        weight_decay_penalty: float
            the magnitude of the weight decay penalty to use
        weight_decay_penalty_type: str
            the type of penalty to use for weight decay, either 'l1' or 'l2'
        dropouts: list or float
            the dropout probablity to use for each layer.  The length of this list should equal len(layer_sizes).
            Alternatively this may be a single value instead of a list, in which case the same value is used for every layer.
        activation_fns: list or object
            the Tensorflow activation function to apply to each layer.  The length of this list should equal
            len(layer_sizes).  Alternatively this may be a single value instead of a list, in which case the
            same value is used for every layer.
        """

        if not isinstance(activation_fns, str):
            logger.warning(
                "Activation function should be a string matching an attribute in torch.nn.functional corresponding to an activation function. Using 'relu' as default."
            )
            activation_fns = 'relu'

        super().__init__(n_tasks=n_tasks,
                         n_features=n_features,
                         layer_sizes=layer_sizes,
                         mode='classification',
                         alpha_init_stddevs=alpha_init_stddevs,
                         weight_init_stddevs=weight_init_stddevs,
                         bias_init_consts=bias_init_consts,
                         weight_decay_penalty=weight_decay_penalty,
                         weight_decay_penalty_type=weight_decay_penalty_type,
                         activation_fns=activation_fns,
                         dropouts=dropouts,
                         n_outputs=n_outputs,
                         n_classes=n_classes,
                         **kwargs)



import ctypes
import ctypes.util
import dqclibs
import numpy as np

# CONSTANTS
NDIM = 3

CINT = dqclibs.CINT
CGTO = dqclibs.CGTO
CPBC = dqclibs.CPBC
# CVHF = dqclibs.CVHF
CSYMM = dqclibs.CSYMM

c_null_ptr = ctypes.POINTER(ctypes.c_void_p)


def np2ctypes(a: np.ndarray) -> ctypes.c_void_p:
    """Get the ctypes of the numpy ndarray

    Examples
    --------
    >>> import numpy as np
    >>> a = np.array([1, 2, 3])
    >>> ctype = np2ctypes(a)

    Parameters
    ----------
    a : np.ndarray
        Numpy ndarray

    Returns
    -------
    ctypes.c_void_p
        ctypes of the numpy ndarray

    """
    return a.ctypes.data_as(ctypes.c_void_p)


def int2ctypes(a: int) -> ctypes.c_int:
    """Convert the python's integer to ctypes' integer

    Examples
    --------
    >>> a = 1
    >>> ctype = int2ctypes(a)

    Parameters
    ----------
    a : int
        Python's integer

    Returns
    -------
    ctypes.c_int
        ctypes' integer

    """
    return ctypes.c_int(a)



import math
try:
    import torch
    from torch.optim import Optimizer
except:
    pass
from deepchem.utils.typing import Betas2, Params, Callable, Any
from typing import Optional


class LambOptimizer(Optimizer):
    r"""Implements Lamb algorithm.

    It has been proposed in `Large Batch Optimization for Deep Learning:
    Training BERT in 76 minutes`__.

    Parameters
    ----------
    params: Params
        iterable of parameters to optimize or dicts defining parameter groups
    lr: float
        learning rate (default: 1e-3)
    betas: Betas2
        coefficients used for computing running averages of gradient and its square (default: (0.9, 0.999))
    eps: float
        term added to the denominator to improve numerical stability (default: 1e-8)
    weight_decay: float
        weight decay (L2 penalty) (default: 0)
    clamp_value: float
        clamp weight_norm in (0,clamp_value) (default: 10) set to a high value to avoid it (e.g 10e3)
    adam: bool
        always use trust ratio = 1, which turns this into Adam. Useful for comparison purposes. (default: False)
    debias: bool
        debias adam by (1 - beta**step) (default: False)

    Example
    -------
    >>> import deepchem as dc
    >>> from deepchem.utils.optimizer_utils import LambOptimizer
    >>> import torch.nn as nn
    >>> model = nn.Sequential(nn.Linear(10, 1))
    >>> input = torch.randn((5, 10))
    >>> target = torch.randn((5, 1))
    >>> loss_fn = nn.MSELoss()
    >>> optimizer = LambOptimizer(model.parameters(), lr=0.1)
    >>> optimizer.zero_grad()
    >>> loss_fn(model(input), target).backward()
    >>> optimizer.step()

    __ https://arxiv.org/abs/1904.00962

    Note
    ----
    Reference code: https://github.com/cybertronai/pytorch-lamb
    """

    def __init__(
        self,
        params: Params,
        lr: float = 1e-3,
        betas: Betas2 = (0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0,
        clamp_value: float = 10,
        adam: bool = False,
        debias: bool = False,
    ) -> None:
        if lr <= 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if eps < 0.0:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(
                betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(
                betas[1]))
        if weight_decay < 0:
            raise ValueError(
                "Invalid weight_decay value: {}".format(weight_decay))
        if clamp_value < 0.0:
            raise ValueError("Invalid clamp value: {}".format(clamp_value))

        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay)
        self.clamp_value = clamp_value
        self.adam = adam
        self.debias = debias

        super(LambOptimizer, self).__init__(params, defaults)

    def step(self, closure: Optional[Callable[[], float]] = None) -> Any:
        r"""Performs a single optimization step.

        Arguments:
            closure: A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    msg = ("Lamb does not support sparse gradients, "
                           "please consider SparseAdam instead")
                    raise RuntimeError(msg)

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(
                        p, memory_format=torch.preserve_format)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                # m_t
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                # v_t
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1 - beta2)

                # Paper v3 does not use debiasing.
                if self.debias:
                    bias_correction = math.sqrt(1 - beta2**state["step"])
                    bias_correction /= 1 - beta1**state["step"]
                else:
                    bias_correction = 1

                # Apply bias to lr to avoid broadcast.
                step_size = group["lr"] * bias_correction

                weight_norm = torch.norm(p.data).clamp(0, self.clamp_value)

                adam_step = exp_avg / exp_avg_sq.sqrt().add(group["eps"])
                if group["weight_decay"] != 0:
                    adam_step.add_(p.data, alpha=group["weight_decay"])

                adam_norm = torch.norm(adam_step)
                if weight_norm == 0 or adam_norm == 0:
                    trust_ratio = 1
                else:
                    trust_ratio = weight_norm / adam_norm
                state["weight_norm"] = weight_norm
                state["adam_norm"] = adam_norm
                state["trust_ratio"] = trust_ratio
                if self.adam:
                    trust_ratio = 1

                p.data.add_(adam_step, alpha=-step_size * trust_ratio)

        return loss
